# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_docs.ipynb.

# %% ../nbs/00_docs.ipynb 3
from __future__ import annotations
from .import_essentials import *
import nbdev
from fastcore.basics import AttrDict
from fastcore.utils import *

from nbdev.showdoc import *
from nbdev.doclinks import *
from inspect import isclass
from nbdev.showdoc import (
    _ext_link, 
    _wrap_sig, 
    _fmt_anno, 
    _f_name, 
    DocmentTbl, 
    _maybe_nm, 
    _show_param
)
from nbdev.config import get_config

# %% auto 0
__all__ = ['CalloutDocument', 'CustomizedMarkdownRenderer']

# %% ../nbs/00_docs.ipynb 4
def _docment_parser(parser: BaseParser):
    p = parser.schema()['properties']
    anno = parser.__annotations__
    d = { 
        k: {
            'anno': anno[k],
            'default': v['default'] if 'default' in v else inspect._empty,
            'docment': v['description'] if 'description' in v else inspect._empty,
        } for k, v in p.items()
    }

    d = AttrDict(d)
    return d


# %% ../nbs/00_docs.ipynb 5
class ParserMarkdownRenderer(BasicMarkdownRenderer):
    def __init__(self, sym, name: str | None = None, title_level: int = 3):
        super().__init__(sym, name, title_level)
        self.dm.dm = _docment_parser(sym)

# %% ../nbs/00_docs.ipynb 6
def _italic(s: str): return f'<em>{s}</em>' if s.strip() else s

def _bold(s: str): return f'<b>{s}</b>' if s.strip() else s

# %% ../nbs/00_docs.ipynb 7
def _show_param(param):
    "Like `Parameter.__str__` except removes: quotes in annos, spaces, ids in reprs"
    kind,res,anno,default = param.kind,param._name,param._annotation,param._default
    kind = '*' if kind==inspect._VAR_POSITIONAL else '**' if kind==inspect._VAR_KEYWORD else ''
    res = kind+res
    # if anno is not inspect._empty: res += f':{_f_name(anno) or _fmt_anno(anno)}'
    if default is not inspect._empty: res += f'={_f_name(default) or repr(default)}'
    return res


def _fmt_sig(sig):
    if sig is None: return ''
    p = {k:v for k,v in sig.parameters.items()}
    _params = [_show_param(p[k]) for k in p.keys() if k != 'self']
    return "(" + ', '.join(_params)  + ")"


# %% ../nbs/00_docs.ipynb 8
def _inner_list2mdlist(l: list):
    param_name, param_anno, param_default, param_doc = l
    # annotation
    if param_anno == inspect._empty: param_anno = None
    else: param_anno = f"`{param_anno}`"
    # default value
    if param_default == inspect._empty: param_default = None
    else: param_default = _italic(f"default={param_default}")

    mdoc = ""
    if param_anno and param_default:
        mdoc += f"* {_bold(param_name)} ({param_anno}, {param_default})"
    elif param_anno:
        mdoc += f"* {_bold(param_name)} ({param_anno})"
    elif param_default:
        mdoc += f"* {_bold(param_name)} ({param_default})"
    else:
        mdoc += f"* {_bold(param_name)}"
    
    if not (param_doc == inspect._empty): 
        mdoc += f" -- {param_doc}"
    return mdoc

def _params_mdlist(tbl: DocmentTbl):
    param_list = [
        L([k, v['anno'], v['default'], v['docment']])
        for k, v in tbl.dm.items() if k != 'return'
    ]
    # param_list = tbl._row_list
    return L(param_list).map(_inner_list2mdlist)

def _return_mdlist(tbl: DocmentTbl):
    return_list = [tbl.dm['return'][k] for k in ['anno', 'default', 'docment']]
    param_anno, param_default, param_doc = return_list
    mdoc = ""
    if not param_anno == inspect._empty: 
        mdoc += f"(`{param_anno}`)"
    if param_doc != inspect._empty:
        mdoc += f" -- {param_doc}"
    return mdoc

def _show_params_return(tbl: DocmentTbl):
    if not tbl.has_docment: return ''
    doc = "" 
    doc = "::: {#docs .callout-note icon=false}\n\n"
    doc += '## Parameters:' + '\n'
    doc += _params_mdlist(tbl)
    doc += "\n\n:::\n\n"
    if tbl.has_return:
        doc += "::: {#docs .callout-note icon=false}\n\n"
        doc += '\n\n## Returns:\n'
        doc += f"&ensp;&ensp;&ensp;&ensp;{_return_mdlist(tbl)}"
        doc += "\n\n:::"
    
    return '\n'.join(doc)

# %% ../nbs/00_docs.ipynb 9
class CalloutDocument():
    def __init__(self, tbl: DocmentTbl):
        self.tbl = tbl
    
    def _repre_mardown(self):
        return _show_params_return(self.tbl)

    __str__ = _repre_mardown

# %% ../nbs/00_docs.ipynb 10
class CustomizedMarkdownRenderer(ShowDocRenderer):
    """Displaying documents of functions, classes, `haiku.module`, and `BaseParser`."""
    
    def __init__(self, sym, name:str|None=None, title_level:int=3):
        super().__init__(sym, name, title_level)
        self.isclass = inspect.isclass(sym)
        self._check_sym(sym)

    def _check_sym(self, sym):
       
        if self.isclass:
            # extract annotations for pydantic models
            if issubclass(sym, BaseParser):
                self.dm.dm = _docment_parser(sym)
            # extract annotations for hk.Module
            elif issubclass(sym, hk.Module):
                _sym = sym.__init__
                try: self.sig = signature_ex(_sym, eval_str=True)
                except (ValueError,TypeError): self.sig = None
                self.dm = DocmentTbl(_sym)

    def _repr_markdown_(self):
        doc = '---\n\n'
        # doc = ''
        src = NbdevLookup().code(self.fn)
        _look_up = NbdevLookup()[self.fn]
        if _look_up: 
            module_dir = _look_up[1].replace('.py', '').replace('/', '.') + '.'
        else:
            module_dir = ""
        
        if src: 
            link = _ext_link(src, 'source', 'style="float:right; font-size:smaller"') + '\n\n'
        else:
            link = ''
        doc += link
        h = '#'*self.title_level
        doc += f'{h} {str(self.nm).upper()}\n\n'
        # if self.isclass: doc += '> *class* '
        # else: doc += '> '
        if self.isclass: doc += '::: {.doc-sig}\n\n CLASS '
        else: doc += '::: {.doc-sig}\n\n '
        sig = f"{module_dir}{_bold(self.nm)} {_italic(_fmt_sig(self.sig))}\n\n:::"
        doc += f'{sig}'
        if self.docs: doc += f"\n\n{self.docs}"
        # if self.dm.has_docment: doc += f"\n\n{_show_params_return(self.dm)}"
        if self.dm.has_docment: doc += f"\n\n{CalloutDocument(self.dm)}"

        return doc




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_data.scm.ipynb.

# %% ../../nbs/01_data.scm.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from . import TabularDataModule, TabularDataModuleConfigs
import networkx as nx
try: 
    from causalgraphicalmodels import CausalGraphicalModel, StructuralCausalModel
except: 
    raise ModuleNotFoundError("`causalgraphicalmodels` is not installed. "
        "Try to install via `pip install git+https://github.com/BirkhoffG/causalgraphicalmodels.git`.")

# %% auto 0
__all__ = ['BaseDistribution', 'Normal', 'MixtureOfGaussians', 'sanity_3_lin', 'CausalModel']

# %% ../../nbs/01_data.scm.ipynb 5
class BaseDistribution(ABC):
    """Base class for all distributions."""
    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def sample(self, size: int):
        pass

    @abstractmethod
    def pdf(self, value):
        pass

# %% ../../nbs/01_data.scm.ipynb 6
class Normal(BaseDistribution):
    """Normal distribution."""
    def __init__(self, mean: int | float, var: int | float):
        super().__init__(f"Normal (mean={mean}, var={var})")
        self.mean = mean
        self.var = var

    def sample(self, size=1):
        tmp = [np.random.normal(self.mean, np.sqrt(self.var)) for _ in range(size)]
        return tmp[0] if size == 1 else tmp

    def pdf(self, value):
        return scipy.stats.norm(self.mean, self.var).pdf(value)


class MixtureOfGaussians(BaseDistribution):
    """Mixture of Gaussians distribution."""
    def __init__(self, probs, means, vars):
        if not np.sum(probs) == 1:
            raise ValueError("Mixture probabilities must sum to 1.")
        if not len(probs) == len(means) == len(vars):
            raise ValueError("Length mismatch.")

        self.probs = probs
        self.means = means
        self.vars = vars
        super().__init__(
            f"MixtureOfGaussians (mean={probs}, means={means}, vars={vars}"
        )

    def sample(self, size=1):
        tmp = [
            np.random.normal(self.means[mixture_idx], np.sqrt(self.vars[mixture_idx]))
            for mixture_idx in np.random.choice(
                len(self.probs), size=size, p=self.probs
            )
        ]
        return tmp[0] if size == 1 else tmp

    def pdf(self, value):
        return np.sum(
            [
                prob * scipy.stats.norm(mean, var).pdf(value)
                for (prob, mean, var) in zip(self.probs, self.means, self.vars)
            ]
        )


# %% ../../nbs/01_data.scm.ipynb 8
def sanity_3_lin():
    structural_equations = {
        "x1": lambda n_samples: n_samples,
        "x2": lambda n_samples, x1: -x1 + n_samples,
        "x3": lambda n_samples, x1, x2: 0.5 * (0.1 * x1 + 0.5 * x2) + n_samples,
    }
    noises_distributions = {
        "u1": MixtureOfGaussians([0.5, 0.5], [-2, +1], [1.5, 1]),
        "u2": Normal(0, 1),
        "u3": Normal(0, 1),
    }
    continuous = list(structural_equations.keys()) + list(
        noises_distributions.keys()
    )
    categorical = []
    immutables = []

    return (
        structural_equations,
        noises_distributions,
        continuous,
        categorical,
        immutables,
    )

# %% ../../nbs/01_data.scm.ipynb 9
def _load_scm_equations(
    scm_class: str # Name of the structural causal model.
) -> Tuple[
    Dict[str, Callable], 
    Dict[str, Callable], 
    list, list, list]: # return scm equations
    
    scm_func_dict = {
        'sanity_3_lin': sanity_3_lin
    }
    if not scm_class in scm_func_dict:
        raise ValueError(f'{scm_class} should be one of {scm_func_dict.keys()}')
    return scm_func_dict[scm_class]()

# %% ../../nbs/01_data.scm.ipynb 11
class CausalModel:
    """Class with topological methods given a structural causal model."""
    def __init__(
        self,
        scm_class: str # Name of the structural causal model.
    ):
        (
            self._structural_equations,
            self._noise_distributions,
            self._continuous,
            self._categorical,
            self._immutables,
        ) = _load_scm_equations(scm_class)
        self._scm = StructuralCausalModel(self._structural_equations)
        self._cgm = self._scm.cgm

        self._endogenous = list(self._structural_equations.keys())
        self._exogenous = list(self._noise_distributions.keys())

        self._continuous_noise = list(set(self._continuous) - set(self._endogenous))
        self._categorical_noise = list(set(self._categorical) - set(self._endogenous))

        self._continuous = list(set(self._continuous) - set(self._exogenous))
        self._categorical = list(set(self._categorical) - set(self._exogenous))

    def get_topological_ordering(self, node_type: str = 'endogenous'):
        """Get topological order of nodes."""
        tmp = nx.topological_sort(self._cgm.dag)
        if node_type == 'endogenous':
            return tmp
        elif node_type == 'exogenous':
            return [f'u{node[1:]}' for node in tmp]
        else:
            raise ValueError(f'node_type should be one of ["endogenous", "exogenous"], '
                             f'but got {node_type}')
        
    def get_children(self, node: str):
        """Get children of a node."""
        return set(self._cgm.dag.successors(node))
    
    def get_parents(self, node: str, return_sorted: bool = True):
        """Get parents of a node."""
        tmp = set(self._cgm.dag.predecessors(node))
        return sorted(tmp) if return_sorted else tmp
    
    def get_ancestors(self, node: str, return_sorted: bool = True):
        """Get ancestors of a node."""
        tmp = nx.ancestors(self._cgm.dag, node)
        return sorted(tmp) if return_sorted else tmp
    
    def get_descendants(self, node: str, return_sorted: bool = True):
        """Get descendants of a node."""
        tmp = nx.descendants(self._cgm.dag, node)
        return sorted(tmp) if return_sorted else tmp
    
    def get_non_descendants(self, node: str):
        """Get non-descendants of a node."""
        return (
            set(nx.topological_sort(self._cgm.dag))
            .difference(self.get_descendants(node))
            .symmetric_difference(set([node]))
        )
    
    def visualize(self, dir_name: str = None):
        """Visualize the causal graph."""
        if dir_name:
            path = f'{dir_name}/_causal_graph'
            view_flag = False
        else:
            path = "_tmp/_causal_graph"
            view_flag = True
        self._cgm.draw().render(path, view=view_flag)

    @property
    def scm(self) -> StructuralCausalModel:
        return self._scm

    @property
    def cgm(self) -> CausalGraphicalModel:
        return self._cgm

    @property
    def scm_class(self) -> str:
        return self._scm_class

    @property
    def structural_equations(self) -> dict:
        return self._structural_equations

    @property
    def noise_distributions(self) -> dict:
        return self._noise_distributions

    @property
    def exogenous(self) -> List[str]:
        return self._exogenous

    @property
    def endogenous(self) -> List[str]:
        return self._endogenous

# %% ../../nbs/01_data.scm.ipynb 16
def _get_noise_string(node):
    if not node[0] == "x":
        raise ValueError
    return "u" + node[1:] # get node id

# %% ../../nbs/01_data.scm.ipynb 17
def _create_synthetic_data(
    scm: CausalModel, # Structural causal model 
    num_samples: int # Number of samples in the dataset.
) -> Union[pd.DataFrame, pd.DataFrame]: # Return exogenous and endogenous variables.
    """Generate synthetic data."""

    exogenous_nodes = scm.get_topological_ordering("exogenous")
    exogenous_variables = np.concatenate(
        [
            np.array(scm.noise_distributions[node].sample(num_samples)).reshape(-1, 1)
            for node in exogenous_nodes
        ],
        axis=1,
    )
    exogenous_variables = pd.DataFrame(
        exogenous_variables, columns=exogenous_nodes
    )

    endogenous_variables = exogenous_variables.copy()
    endogenous_variables = endogenous_variables.rename(
        columns=dict(
            zip(
                scm.get_topological_ordering("exogenous"),
                scm.get_topological_ordering("endogenous"),
            )
        )
    )
    # used later to make sure parents are populated when computing children
    endogenous_variables.loc[:] = np.nan
    for node in scm.get_topological_ordering("endogenous"):
        parents = scm.get_parents(node)
        if endogenous_variables.loc[:, list(parents)].isnull().values.any():
            raise ValueError(
                "parents in endogenous_variables should already be occupied"
            )
        endogenous_variables[node] = scm.structural_equations[node](
            exogenous_variables[_get_noise_string(node)],
            *[endogenous_variables[p] for p in parents],
        )

    # fix a hyperplane
    w = np.ones((endogenous_variables.shape[1], 1))
    # get the average scale of (w^T)*X, this depends on the scale of the data
    scale = 2.5 / np.mean(np.abs(np.dot(endogenous_variables, w)))
    predictions = 1 / (1 + np.exp(-scale * np.dot(endogenous_variables, w)))

    if not 0.20 < np.std(predictions) < 0.42:
        raise ValueError(f"std of labels is strange: {np.std(predictions)}")

    # sample labels from class probabilities in predictions
    uniform_rv = np.random.rand(endogenous_variables.shape[0], 1)
    labels = uniform_rv < predictions
    labels = pd.DataFrame(data=labels, columns=["label"])

    df_endogenous = pd.concat([endogenous_variables, labels], axis=1).astype("float64")
    df_exogenous = pd.concat([exogenous_variables], axis=1).astype("float64")
    return df_endogenous, df_exogenous



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/ReLax/',
                'doc_host': 'https://birkhoffg.github.io',
                'git_url': 'https://github.com/birkhoffg/relax/tree/master/',
                'lib_path': 'relax'},
  'syms': { 'relax.data.loader': { 'relax.data.loader.ArrayDataset': ('data.loader.html#arraydataset', 'relax/data/loader.py'),
                                   'relax.data.loader.ArrayDataset.__getitem__': ( 'data.loader.html#arraydataset.__getitem__',
                                                                                   'relax/data/loader.py'),
                                   'relax.data.loader.ArrayDataset.__init__': ( 'data.loader.html#arraydataset.__init__',
                                                                                'relax/data/loader.py'),
                                   'relax.data.loader.ArrayDataset.__len__': ( 'data.loader.html#arraydataset.__len__',
                                                                               'relax/data/loader.py'),
                                   'relax.data.loader.BaseDataLoader': ('data.loader.html#basedataloader', 'relax/data/loader.py'),
                                   'relax.data.loader.BaseDataLoader.__init__': ( 'data.loader.html#basedataloader.__init__',
                                                                                  'relax/data/loader.py'),
                                   'relax.data.loader.BaseDataLoader.__iter__': ( 'data.loader.html#basedataloader.__iter__',
                                                                                  'relax/data/loader.py'),
                                   'relax.data.loader.BaseDataLoader.__len__': ( 'data.loader.html#basedataloader.__len__',
                                                                                 'relax/data/loader.py'),
                                   'relax.data.loader.BaseDataLoader.__next__': ( 'data.loader.html#basedataloader.__next__',
                                                                                  'relax/data/loader.py'),
                                   'relax.data.loader.DataLoader': ('data.loader.html#dataloader', 'relax/data/loader.py'),
                                   'relax.data.loader.DataLoader.__init__': ( 'data.loader.html#dataloader.__init__',
                                                                              'relax/data/loader.py'),
                                   'relax.data.loader.DataloaderBackends': ('data.loader.html#dataloaderbackends', 'relax/data/loader.py'),
                                   'relax.data.loader.DataloaderBackends.__getitem__': ( 'data.loader.html#dataloaderbackends.__getitem__',
                                                                                         'relax/data/loader.py'),
                                   'relax.data.loader.DataloaderBackends.supported': ( 'data.loader.html#dataloaderbackends.supported',
                                                                                       'relax/data/loader.py'),
                                   'relax.data.loader.Dataset': ('data.loader.html#dataset', 'relax/data/loader.py'),
                                   'relax.data.loader.Dataset.__getitem__': ( 'data.loader.html#dataset.__getitem__',
                                                                              'relax/data/loader.py'),
                                   'relax.data.loader.JaxDataloader': ('data.loader.html#jaxdataloader', 'relax/data/loader.py'),
                                   'relax.data.loader.JaxDataloader.__init__': ( 'data.loader.html#jaxdataloader.__init__',
                                                                                 'relax/data/loader.py'),
                                   'relax.data.loader.JaxDataloader.__iter__': ( 'data.loader.html#jaxdataloader.__iter__',
                                                                                 'relax/data/loader.py'),
                                   'relax.data.loader.JaxDataloader.__len__': ( 'data.loader.html#jaxdataloader.__len__',
                                                                                'relax/data/loader.py'),
                                   'relax.data.loader.JaxDataloader.__next__': ( 'data.loader.html#jaxdataloader.__next__',
                                                                                 'relax/data/loader.py'),
                                   'relax.data.loader.JaxDataloader._shuffle': ( 'data.loader.html#jaxdataloader._shuffle',
                                                                                 'relax/data/loader.py'),
                                   'relax.data.loader.JaxDataloader._stop_iteration': ( 'data.loader.html#jaxdataloader._stop_iteration',
                                                                                        'relax/data/loader.py'),
                                   'relax.data.loader.TorchDataloader': ('data.loader.html#torchdataloader', 'relax/data/loader.py'),
                                   'relax.data.loader.TorchDataloader.__init__': ( 'data.loader.html#torchdataloader.__init__',
                                                                                   'relax/data/loader.py'),
                                   'relax.data.loader.TorchDataloader.__iter__': ( 'data.loader.html#torchdataloader.__iter__',
                                                                                   'relax/data/loader.py'),
                                   'relax.data.loader.TorchDataloader.__len__': ( 'data.loader.html#torchdataloader.__len__',
                                                                                  'relax/data/loader.py'),
                                   'relax.data.loader.TorchDataloader.__next__': ( 'data.loader.html#torchdataloader.__next__',
                                                                                   'relax/data/loader.py'),
                                   'relax.data.loader._convert_dataset_pytorch': ( 'data.loader.html#_convert_dataset_pytorch',
                                                                                   'relax/data/loader.py'),
                                   'relax.data.loader._dispatch_dataloader': ( 'data.loader.html#_dispatch_dataloader',
                                                                               'relax/data/loader.py'),
                                   'relax.data.loader._numpy_collate': ('data.loader.html#_numpy_collate', 'relax/data/loader.py')},
            'relax.data.module': { 'relax.data.module.BaseDataModule': ('data.module.html#basedatamodule', 'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.apply_constraints': ( 'data.module.html#basedatamodule.apply_constraints',
                                                                                           'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.apply_regularization': ( 'data.module.html#basedatamodule.apply_regularization',
                                                                                              'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.data': ( 'data.module.html#basedatamodule.data',
                                                                              'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.data_name': ( 'data.module.html#basedatamodule.data_name',
                                                                                   'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.dataset': ( 'data.module.html#basedatamodule.dataset',
                                                                                 'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.inverse_transform': ( 'data.module.html#basedatamodule.inverse_transform',
                                                                                           'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.prepare_data': ( 'data.module.html#basedatamodule.prepare_data',
                                                                                      'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.test_dataloader': ( 'data.module.html#basedatamodule.test_dataloader',
                                                                                         'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.test_dataset': ( 'data.module.html#basedatamodule.test_dataset',
                                                                                      'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.train_dataloader': ( 'data.module.html#basedatamodule.train_dataloader',
                                                                                          'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.train_dataset': ( 'data.module.html#basedatamodule.train_dataset',
                                                                                       'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.transform': ( 'data.module.html#basedatamodule.transform',
                                                                                   'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.val_dataloader': ( 'data.module.html#basedatamodule.val_dataloader',
                                                                                        'relax/data/module.py'),
                                   'relax.data.module.BaseDataModule.val_dataset': ( 'data.module.html#basedatamodule.val_dataset',
                                                                                     'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule': ('data.module.html#tabulardatamodule', 'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.__init__': ( 'data.module.html#tabulardatamodule.__init__',
                                                                                     'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.__setattr__': ( 'data.module.html#tabulardatamodule.__setattr__',
                                                                                        'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.apply_constraints': ( 'data.module.html#tabulardatamodule.apply_constraints',
                                                                                              'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.apply_regularization': ( 'data.module.html#tabulardatamodule.apply_regularization',
                                                                                                 'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.data': ( 'data.module.html#tabulardatamodule.data',
                                                                                 'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.data_name': ( 'data.module.html#tabulardatamodule.data_name',
                                                                                      'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.dataset': ( 'data.module.html#tabulardatamodule.dataset',
                                                                                    'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.inverse_transform': ( 'data.module.html#tabulardatamodule.inverse_transform',
                                                                                              'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.prepare_data': ( 'data.module.html#tabulardatamodule.prepare_data',
                                                                                         'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.test_dataloader': ( 'data.module.html#tabulardatamodule.test_dataloader',
                                                                                            'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.test_dataset': ( 'data.module.html#tabulardatamodule.test_dataset',
                                                                                         'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.train_dataloader': ( 'data.module.html#tabulardatamodule.train_dataloader',
                                                                                             'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.train_dataset': ( 'data.module.html#tabulardatamodule.train_dataset',
                                                                                          'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.transform': ( 'data.module.html#tabulardatamodule.transform',
                                                                                      'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.val_dataloader': ( 'data.module.html#tabulardatamodule.val_dataloader',
                                                                                           'relax/data/module.py'),
                                   'relax.data.module.TabularDataModule.val_dataset': ( 'data.module.html#tabulardatamodule.val_dataset',
                                                                                        'relax/data/module.py'),
                                   'relax.data.module.TabularDataModuleConfigs': ( 'data.module.html#tabulardatamoduleconfigs',
                                                                                   'relax/data/module.py'),
                                   'relax.data.module.TabularDataModuleConfigs.Config': ( 'data.module.html#tabulardatamoduleconfigs.config',
                                                                                          'relax/data/module.py'),
                                   'relax.data.module.TransformerMixinType': ( 'data.module.html#transformermixintype',
                                                                               'relax/data/module.py'),
                                   'relax.data.module.TransformerMixinType.__get_validators__': ( 'data.module.html#transformermixintype.__get_validators__',
                                                                                                  'relax/data/module.py'),
                                   'relax.data.module.TransformerMixinType.__modify_schema__': ( 'data.module.html#transformermixintype.__modify_schema__',
                                                                                                 'relax/data/module.py'),
                                   'relax.data.module.TransformerMixinType.validate': ( 'data.module.html#transformermixintype.validate',
                                                                                        'relax/data/module.py'),
                                   'relax.data.module._check_cols': ('data.module.html#_check_cols', 'relax/data/module.py'),
                                   'relax.data.module._init_scalar_encoder': ( 'data.module.html#_init_scalar_encoder',
                                                                               'relax/data/module.py'),
                                   'relax.data.module._inverse_transform_np': ( 'data.module.html#_inverse_transform_np',
                                                                                'relax/data/module.py'),
                                   'relax.data.module._process_data': ('data.module.html#_process_data', 'relax/data/module.py'),
                                   'relax.data.module._supported_backends': ( 'data.module.html#_supported_backends',
                                                                              'relax/data/module.py'),
                                   'relax.data.module._transform_df': ('data.module.html#_transform_df', 'relax/data/module.py'),
                                   'relax.data.module._validate_dataname': ('data.module.html#_validate_dataname', 'relax/data/module.py'),
                                   'relax.data.module.find_imutable_idx_list': ( 'data.module.html#find_imutable_idx_list',
                                                                                 'relax/data/module.py'),
                                   'relax.data.module.load_data': ('data.module.html#load_data', 'relax/data/module.py'),
                                   'relax.data.module.sample': ('data.module.html#sample', 'relax/data/module.py')},
            'relax.data.scm': { 'relax.data.scm.BaseDistribution': ('data.scm.html#basedistribution', 'relax/data/scm.py'),
                                'relax.data.scm.BaseDistribution.__init__': ( 'data.scm.html#basedistribution.__init__',
                                                                              'relax/data/scm.py'),
                                'relax.data.scm.BaseDistribution.pdf': ('data.scm.html#basedistribution.pdf', 'relax/data/scm.py'),
                                'relax.data.scm.BaseDistribution.sample': ('data.scm.html#basedistribution.sample', 'relax/data/scm.py'),
                                'relax.data.scm.CausalModel': ('data.scm.html#causalmodel', 'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.__init__': ('data.scm.html#causalmodel.__init__', 'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.cgm': ('data.scm.html#causalmodel.cgm', 'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.endogenous': ('data.scm.html#causalmodel.endogenous', 'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.exogenous': ('data.scm.html#causalmodel.exogenous', 'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.get_ancestors': ( 'data.scm.html#causalmodel.get_ancestors',
                                                                              'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.get_children': ('data.scm.html#causalmodel.get_children', 'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.get_descendants': ( 'data.scm.html#causalmodel.get_descendants',
                                                                                'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.get_non_descendants': ( 'data.scm.html#causalmodel.get_non_descendants',
                                                                                    'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.get_parents': ('data.scm.html#causalmodel.get_parents', 'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.get_topological_ordering': ( 'data.scm.html#causalmodel.get_topological_ordering',
                                                                                         'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.noise_distributions': ( 'data.scm.html#causalmodel.noise_distributions',
                                                                                    'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.scm': ('data.scm.html#causalmodel.scm', 'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.scm_class': ('data.scm.html#causalmodel.scm_class', 'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.structural_equations': ( 'data.scm.html#causalmodel.structural_equations',
                                                                                     'relax/data/scm.py'),
                                'relax.data.scm.CausalModel.visualize': ('data.scm.html#causalmodel.visualize', 'relax/data/scm.py'),
                                'relax.data.scm.MixtureOfGaussians': ('data.scm.html#mixtureofgaussians', 'relax/data/scm.py'),
                                'relax.data.scm.MixtureOfGaussians.__init__': ( 'data.scm.html#mixtureofgaussians.__init__',
                                                                                'relax/data/scm.py'),
                                'relax.data.scm.MixtureOfGaussians.pdf': ('data.scm.html#mixtureofgaussians.pdf', 'relax/data/scm.py'),
                                'relax.data.scm.MixtureOfGaussians.sample': ( 'data.scm.html#mixtureofgaussians.sample',
                                                                              'relax/data/scm.py'),
                                'relax.data.scm.Normal': ('data.scm.html#normal', 'relax/data/scm.py'),
                                'relax.data.scm.Normal.__init__': ('data.scm.html#normal.__init__', 'relax/data/scm.py'),
                                'relax.data.scm.Normal.pdf': ('data.scm.html#normal.pdf', 'relax/data/scm.py'),
                                'relax.data.scm.Normal.sample': ('data.scm.html#normal.sample', 'relax/data/scm.py'),
                                'relax.data.scm._create_synthetic_data': ('data.scm.html#_create_synthetic_data', 'relax/data/scm.py'),
                                'relax.data.scm._get_noise_string': ('data.scm.html#_get_noise_string', 'relax/data/scm.py'),
                                'relax.data.scm._load_scm_equations': ('data.scm.html#_load_scm_equations', 'relax/data/scm.py'),
                                'relax.data.scm.sanity_3_lin': ('data.scm.html#sanity_3_lin', 'relax/data/scm.py')},
            'relax.docs': { 'relax.docs.CalloutDocument': ('docs.html#calloutdocument', 'relax/docs.py'),
                            'relax.docs.CalloutDocument.__init__': ('docs.html#calloutdocument.__init__', 'relax/docs.py'),
                            'relax.docs.CalloutDocument._repre_mardown': ('docs.html#calloutdocument._repre_mardown', 'relax/docs.py'),
                            'relax.docs.CustomizedMarkdownRenderer': ('docs.html#customizedmarkdownrenderer', 'relax/docs.py'),
                            'relax.docs.CustomizedMarkdownRenderer.__init__': ( 'docs.html#customizedmarkdownrenderer.__init__',
                                                                                'relax/docs.py'),
                            'relax.docs.CustomizedMarkdownRenderer._check_sym': ( 'docs.html#customizedmarkdownrenderer._check_sym',
                                                                                  'relax/docs.py'),
                            'relax.docs.CustomizedMarkdownRenderer._repr_markdown_': ( 'docs.html#customizedmarkdownrenderer._repr_markdown_',
                                                                                       'relax/docs.py'),
                            'relax.docs.ParserMarkdownRenderer': ('docs.html#parsermarkdownrenderer', 'relax/docs.py'),
                            'relax.docs.ParserMarkdownRenderer.__init__': ('docs.html#parsermarkdownrenderer.__init__', 'relax/docs.py'),
                            'relax.docs._bold': ('docs.html#_bold', 'relax/docs.py'),
                            'relax.docs._docment_parser': ('docs.html#_docment_parser', 'relax/docs.py'),
                            'relax.docs._fmt_sig': ('docs.html#_fmt_sig', 'relax/docs.py'),
                            'relax.docs._inner_list2mdlist': ('docs.html#_inner_list2mdlist', 'relax/docs.py'),
                            'relax.docs._italic': ('docs.html#_italic', 'relax/docs.py'),
                            'relax.docs._params_mdlist': ('docs.html#_params_mdlist', 'relax/docs.py'),
                            'relax.docs._return_mdlist': ('docs.html#_return_mdlist', 'relax/docs.py'),
                            'relax.docs._show_param': ('docs.html#_show_param', 'relax/docs.py'),
                            'relax.docs._show_params_return': ('docs.html#_show_params_return', 'relax/docs.py')},
            'relax.evaluate': { 'relax.evaluate.BaseEvalMetrics': ('evaluate.html#baseevalmetrics', 'relax/evaluate.py'),
                                'relax.evaluate.BaseEvalMetrics.__call__': ('evaluate.html#baseevalmetrics.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.BaseEvalMetrics.__init__': ('evaluate.html#baseevalmetrics.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.BaseEvalMetrics.__str__': ('evaluate.html#baseevalmetrics.__str__', 'relax/evaluate.py'),
                                'relax.evaluate.BaseGenerationStrategy': ('evaluate.html#basegenerationstrategy', 'relax/evaluate.py'),
                                'relax.evaluate.BaseGenerationStrategy.__call__': ( 'evaluate.html#basegenerationstrategy.__call__',
                                                                                    'relax/evaluate.py'),
                                'relax.evaluate.Explanation': ('evaluate.html#explanation', 'relax/evaluate.py'),
                                'relax.evaluate.Explanation.__post_init__': ( 'evaluate.html#explanation.__post_init__',
                                                                              'relax/evaluate.py'),
                                'relax.evaluate.IterativeGenerationStrategy': ( 'evaluate.html#iterativegenerationstrategy',
                                                                                'relax/evaluate.py'),
                                'relax.evaluate.IterativeGenerationStrategy.__call__': ( 'evaluate.html#iterativegenerationstrategy.__call__',
                                                                                         'relax/evaluate.py'),
                                'relax.evaluate.ManifoldDist': ('evaluate.html#manifolddist', 'relax/evaluate.py'),
                                'relax.evaluate.ManifoldDist.__call__': ('evaluate.html#manifolddist.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.ManifoldDist.__init__': ('evaluate.html#manifolddist.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.PmapGenerationStrategy': ('evaluate.html#pmapgenerationstrategy', 'relax/evaluate.py'),
                                'relax.evaluate.PmapGenerationStrategy.__call__': ( 'evaluate.html#pmapgenerationstrategy.__call__',
                                                                                    'relax/evaluate.py'),
                                'relax.evaluate.PmapGenerationStrategy.__init__': ( 'evaluate.html#pmapgenerationstrategy.__init__',
                                                                                    'relax/evaluate.py'),
                                'relax.evaluate.PredictiveAccuracy': ('evaluate.html#predictiveaccuracy', 'relax/evaluate.py'),
                                'relax.evaluate.PredictiveAccuracy.__call__': ( 'evaluate.html#predictiveaccuracy.__call__',
                                                                                'relax/evaluate.py'),
                                'relax.evaluate.PredictiveAccuracy.__init__': ( 'evaluate.html#predictiveaccuracy.__init__',
                                                                                'relax/evaluate.py'),
                                'relax.evaluate.Proximity': ('evaluate.html#proximity', 'relax/evaluate.py'),
                                'relax.evaluate.Proximity.__call__': ('evaluate.html#proximity.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.Proximity.__init__': ('evaluate.html#proximity.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.Runtime': ('evaluate.html#runtime', 'relax/evaluate.py'),
                                'relax.evaluate.Runtime.__call__': ('evaluate.html#runtime.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.Runtime.__init__': ('evaluate.html#runtime.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.Sparsity': ('evaluate.html#sparsity', 'relax/evaluate.py'),
                                'relax.evaluate.Sparsity.__call__': ('evaluate.html#sparsity.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.Sparsity.__init__': ('evaluate.html#sparsity.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.StrategyFactory': ('evaluate.html#strategyfactory', 'relax/evaluate.py'),
                                'relax.evaluate.StrategyFactory.__init__': ('evaluate.html#strategyfactory.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.StrategyFactory.get_default_strategy': ( 'evaluate.html#strategyfactory.get_default_strategy',
                                                                                         'relax/evaluate.py'),
                                'relax.evaluate.StrategyFactory.get_strategy': ( 'evaluate.html#strategyfactory.get_strategy',
                                                                                 'relax/evaluate.py'),
                                'relax.evaluate.Validity': ('evaluate.html#validity', 'relax/evaluate.py'),
                                'relax.evaluate.Validity.__call__': ('evaluate.html#validity.__call__', 'relax/evaluate.py'),
                                'relax.evaluate.Validity.__init__': ('evaluate.html#validity.__init__', 'relax/evaluate.py'),
                                'relax.evaluate.VmapGenerationStrategy': ('evaluate.html#vmapgenerationstrategy', 'relax/evaluate.py'),
                                'relax.evaluate.VmapGenerationStrategy.__call__': ( 'evaluate.html#vmapgenerationstrategy.__call__',
                                                                                    'relax/evaluate.py'),
                                'relax.evaluate._AuxPredFn': ('evaluate.html#_auxpredfn', 'relax/evaluate.py'),
                                'relax.evaluate._AuxPredFn.__call__': ('evaluate.html#_auxpredfn.__call__', 'relax/evaluate.py'),
                                'relax.evaluate._AuxPredFn.__init__': ('evaluate.html#_auxpredfn.__init__', 'relax/evaluate.py'),
                                'relax.evaluate._check_aux_pred_fn_args': ('evaluate.html#_check_aux_pred_fn_args', 'relax/evaluate.py'),
                                'relax.evaluate._check_pred_fn': ('evaluate.html#_check_pred_fn', 'relax/evaluate.py'),
                                'relax.evaluate._compute_acc': ('evaluate.html#_compute_acc', 'relax/evaluate.py'),
                                'relax.evaluate._compute_manifold_dist': ('evaluate.html#_compute_manifold_dist', 'relax/evaluate.py'),
                                'relax.evaluate._compute_proximity': ('evaluate.html#_compute_proximity', 'relax/evaluate.py'),
                                'relax.evaluate._compute_spar': ('evaluate.html#_compute_spar', 'relax/evaluate.py'),
                                'relax.evaluate._compute_val': ('evaluate.html#_compute_val', 'relax/evaluate.py'),
                                'relax.evaluate._create_second_order_cfs': ('evaluate.html#_create_second_order_cfs', 'relax/evaluate.py'),
                                'relax.evaluate._get_metric': ('evaluate.html#_get_metric', 'relax/evaluate.py'),
                                'relax.evaluate._pad_divisible_X': ('evaluate.html#_pad_divisible_x', 'relax/evaluate.py'),
                                'relax.evaluate._prepare_module': ('evaluate.html#_prepare_module', 'relax/evaluate.py'),
                                'relax.evaluate._train_parametric_module': ('evaluate.html#_train_parametric_module', 'relax/evaluate.py'),
                                'relax.evaluate._validate_configs': ('evaluate.html#_validate_configs', 'relax/evaluate.py'),
                                'relax.evaluate.benchmark_cfs': ('evaluate.html#benchmark_cfs', 'relax/evaluate.py'),
                                'relax.evaluate.compute_so_proximity': ('evaluate.html#compute_so_proximity', 'relax/evaluate.py'),
                                'relax.evaluate.compute_so_sparsity': ('evaluate.html#compute_so_sparsity', 'relax/evaluate.py'),
                                'relax.evaluate.compute_so_validity': ('evaluate.html#compute_so_validity', 'relax/evaluate.py'),
                                'relax.evaluate.evaluate_cfs': ('evaluate.html#evaluate_cfs', 'relax/evaluate.py'),
                                'relax.evaluate.fake_explanations': ('evaluate.html#fake_explanations', 'relax/evaluate.py'),
                                'relax.evaluate.generate_cf_explanations': ('evaluate.html#generate_cf_explanations', 'relax/evaluate.py')},
            'relax.import_essentials': {},
            'relax.logger': { 'relax.logger.TensorboardLogger': ('logger.html#tensorboardlogger', 'relax/logger.py'),
                              'relax.logger.TensorboardLogger.__init__': ('logger.html#tensorboardlogger.__init__', 'relax/logger.py'),
                              'relax.logger.TensorboardLogger.close': ('logger.html#tensorboardlogger.close', 'relax/logger.py'),
                              'relax.logger.TensorboardLogger.get_last_logs': ( 'logger.html#tensorboardlogger.get_last_logs',
                                                                                'relax/logger.py'),
                              'relax.logger.TensorboardLogger.log': ('logger.html#tensorboardlogger.log', 'relax/logger.py'),
                              'relax.logger.TensorboardLogger.log_dict': ('logger.html#tensorboardlogger.log_dict', 'relax/logger.py'),
                              'relax.logger.TensorboardLogger.log_dir': ('logger.html#tensorboardlogger.log_dir', 'relax/logger.py'),
                              'relax.logger.TensorboardLogger.on_epoch_finished': ( 'logger.html#tensorboardlogger.on_epoch_finished',
                                                                                    'relax/logger.py'),
                              'relax.logger.TensorboardLogger.on_epoch_started': ( 'logger.html#tensorboardlogger.on_epoch_started',
                                                                                   'relax/logger.py'),
                              'relax.logger.TensorboardLogger.save_hyperparams': ( 'logger.html#tensorboardlogger.save_hyperparams',
                                                                                   'relax/logger.py')},
            'relax.methods.base': { 'relax.methods.base.BaseCFModule': ('methods.base.html#basecfmodule', 'relax/methods/base.py'),
                                    'relax.methods.base.BaseCFModule.data_module': ( 'methods.base.html#basecfmodule.data_module',
                                                                                     'relax/methods/base.py'),
                                    'relax.methods.base.BaseCFModule.generate_cf': ( 'methods.base.html#basecfmodule.generate_cf',
                                                                                     'relax/methods/base.py'),
                                    'relax.methods.base.BaseCFModule.generate_cfs': ( 'methods.base.html#basecfmodule.generate_cfs',
                                                                                      'relax/methods/base.py'),
                                    'relax.methods.base.BaseCFModule.hook_data_module': ( 'methods.base.html#basecfmodule.hook_data_module',
                                                                                          'relax/methods/base.py'),
                                    'relax.methods.base.BaseCFModule.name': ( 'methods.base.html#basecfmodule.name',
                                                                              'relax/methods/base.py'),
                                    'relax.methods.base.BaseParametricCFModule': ( 'methods.base.html#baseparametriccfmodule',
                                                                                   'relax/methods/base.py'),
                                    'relax.methods.base.BaseParametricCFModule._is_module_trained': ( 'methods.base.html#baseparametriccfmodule._is_module_trained',
                                                                                                      'relax/methods/base.py'),
                                    'relax.methods.base.BaseParametricCFModule.train': ( 'methods.base.html#baseparametriccfmodule.train',
                                                                                         'relax/methods/base.py'),
                                    'relax.methods.base.BasePredFnCFModule': ( 'methods.base.html#basepredfncfmodule',
                                                                               'relax/methods/base.py'),
                                    'relax.methods.base.BasePredFnCFModule.pred_fn': ( 'methods.base.html#basepredfncfmodule.pred_fn',
                                                                                       'relax/methods/base.py')},
            'relax.methods.cchvae': { 'relax.methods.cchvae.CCHVAE': ('methods/cchvae.html#cchvae', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CCHVAE.__init__': ( 'methods/cchvae.html#cchvae.__init__',
                                                                                'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CCHVAE._is_module_trained': ( 'methods/cchvae.html#cchvae._is_module_trained',
                                                                                          'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CCHVAE.generate_cf': ( 'methods/cchvae.html#cchvae.generate_cf',
                                                                                   'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CCHVAE.generate_cfs': ( 'methods/cchvae.html#cchvae.generate_cfs',
                                                                                    'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CCHVAE.train': ('methods/cchvae.html#cchvae.train', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CCHVAEConfigs': ( 'methods/cchvae.html#cchvaeconfigs',
                                                                              'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE': ('methods/cchvae.html#chvae', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.__init__': ( 'methods/cchvae.html#chvae.__init__',
                                                                               'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.__reparameterize': ( 'methods/cchvae.html#chvae.__reparameterize',
                                                                                       'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE._training_step': ( 'methods/cchvae.html#chvae._training_step',
                                                                                     'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.decode': ('methods/cchvae.html#chvae.decode', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.encode': ('methods/cchvae.html#chvae.encode', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.forward': ( 'methods/cchvae.html#chvae.forward',
                                                                              'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.init_net_opt': ( 'methods/cchvae.html#chvae.init_net_opt',
                                                                                   'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.loss': ('methods/cchvae.html#chvae.loss', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.training_step': ( 'methods/cchvae.html#chvae.training_step',
                                                                                    'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAE.validation_step': ( 'methods/cchvae.html#chvae.validation_step',
                                                                                      'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.CHVAEConfigs': ('methods/cchvae.html#chvaeconfigs', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.Decoder': ('methods/cchvae.html#decoder', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.Decoder.__call__': ( 'methods/cchvae.html#decoder.__call__',
                                                                                 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.Decoder.__init__': ( 'methods/cchvae.html#decoder.__init__',
                                                                                 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.Encoder': ('methods/cchvae.html#encoder', 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.Encoder.__call__': ( 'methods/cchvae.html#encoder.__call__',
                                                                                 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae.Encoder.__init__': ( 'methods/cchvae.html#encoder.__init__',
                                                                                 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae._cchvae_generate': ( 'methods/cchvae.html#_cchvae_generate',
                                                                                 'relax/methods/cchvae.py'),
                                      'relax.methods.cchvae._hyper_sphere_coordindates': ( 'methods/cchvae.html#_hyper_sphere_coordindates',
                                                                                           'relax/methods/cchvae.py')},
            'relax.methods.clue': { 'relax.methods.clue.CLUE': ('methods/clue.html#clue', 'relax/methods/clue.py'),
                                    'relax.methods.clue.CLUE.__init__': ('methods/clue.html#clue.__init__', 'relax/methods/clue.py'),
                                    'relax.methods.clue.CLUE._is_module_trained': ( 'methods/clue.html#clue._is_module_trained',
                                                                                    'relax/methods/clue.py'),
                                    'relax.methods.clue.CLUE.generate_cf': ('methods/clue.html#clue.generate_cf', 'relax/methods/clue.py'),
                                    'relax.methods.clue.CLUE.generate_cfs': ( 'methods/clue.html#clue.generate_cfs',
                                                                              'relax/methods/clue.py'),
                                    'relax.methods.clue.CLUE.train': ('methods/clue.html#clue.train', 'relax/methods/clue.py'),
                                    'relax.methods.clue.CLUEConfigs': ('methods/clue.html#clueconfigs', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Decoder': ('methods/clue.html#decoder', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Decoder.__call__': ('methods/clue.html#decoder.__call__', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Decoder.__init__': ('methods/clue.html#decoder.__init__', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Encoder': ('methods/clue.html#encoder', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Encoder.__call__': ('methods/clue.html#encoder.__call__', 'relax/methods/clue.py'),
                                    'relax.methods.clue.Encoder.__init__': ('methods/clue.html#encoder.__init__', 'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat': ('methods/clue.html#vaegausscat', 'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.__init__': ( 'methods/clue.html#vaegausscat.__init__',
                                                                                 'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat._training_step': ( 'methods/clue.html#vaegausscat._training_step',
                                                                                       'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat._update_categorical_info': ( 'methods/clue.html#vaegausscat._update_categorical_info',
                                                                                                 'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.compute_loss': ( 'methods/clue.html#vaegausscat.compute_loss',
                                                                                     'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.decode': ( 'methods/clue.html#vaegausscat.decode',
                                                                               'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.encode': ( 'methods/clue.html#vaegausscat.encode',
                                                                               'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.init_net_opt': ( 'methods/clue.html#vaegausscat.init_net_opt',
                                                                                     'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.sample': ( 'methods/clue.html#vaegausscat.sample',
                                                                               'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.sample_latent': ( 'methods/clue.html#vaegausscat.sample_latent',
                                                                                      'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.sample_prior': ( 'methods/clue.html#vaegausscat.sample_prior',
                                                                                     'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.sample_step': ( 'methods/clue.html#vaegausscat.sample_step',
                                                                                    'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.training_step': ( 'methods/clue.html#vaegausscat.training_step',
                                                                                      'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCat.validation_step': ( 'methods/clue.html#vaegausscat.validation_step',
                                                                                        'relax/methods/clue.py'),
                                    'relax.methods.clue.VAEGaussCatConfigs': ( 'methods/clue.html#vaegausscatconfigs',
                                                                               'relax/methods/clue.py'),
                                    'relax.methods.clue._clue_generate': ('methods/clue.html#_clue_generate', 'relax/methods/clue.py'),
                                    'relax.methods.clue.kl_divergence': ('methods/clue.html#kl_divergence', 'relax/methods/clue.py')},
            'relax.methods.counternet': { 'relax.methods.counternet.CounterNet': ( 'methods/counternet.html#counternet',
                                                                                   'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet.__init__': ( 'methods/counternet.html#counternet.__init__',
                                                                                            'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet._is_module_trained': ( 'methods/counternet.html#counternet._is_module_trained',
                                                                                                      'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet.generate_cf': ( 'methods/counternet.html#counternet.generate_cf',
                                                                                               'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet.generate_cfs': ( 'methods/counternet.html#counternet.generate_cfs',
                                                                                                'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet.pred_fn': ( 'methods/counternet.html#counternet.pred_fn',
                                                                                           'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNet.train': ( 'methods/counternet.html#counternet.train',
                                                                                         'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetConfigs': ( 'methods/counternet.html#counternetconfigs',
                                                                                          'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetModel': ( 'methods/counternet.html#counternetmodel',
                                                                                        'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetModel.__call__': ( 'methods/counternet.html#counternetmodel.__call__',
                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetModel.__init__': ( 'methods/counternet.html#counternetmodel.__init__',
                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetModelConfigs': ( 'methods/counternet.html#counternetmodelconfigs',
                                                                                               'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule': ( 'methods/counternet.html#counternettrainingmodule',
                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.__init__': ( 'methods/counternet.html#counternettrainingmodule.__init__',
                                                                                                          'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule._explainer_step': ( 'methods/counternet.html#counternettrainingmodule._explainer_step',
                                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule._predictor_step': ( 'methods/counternet.html#counternettrainingmodule._predictor_step',
                                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule._training_step': ( 'methods/counternet.html#counternettrainingmodule._training_step',
                                                                                                                'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule._training_step_logs': ( 'methods/counternet.html#counternettrainingmodule._training_step_logs',
                                                                                                                     'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.exp_loss_fn': ( 'methods/counternet.html#counternettrainingmodule.exp_loss_fn',
                                                                                                             'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.forward': ( 'methods/counternet.html#counternettrainingmodule.forward',
                                                                                                         'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.generate_cfs': ( 'methods/counternet.html#counternettrainingmodule.generate_cfs',
                                                                                                              'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.init_net_opt': ( 'methods/counternet.html#counternettrainingmodule.init_net_opt',
                                                                                                              'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.loss_fn_1': ( 'methods/counternet.html#counternettrainingmodule.loss_fn_1',
                                                                                                           'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.loss_fn_2': ( 'methods/counternet.html#counternettrainingmodule.loss_fn_2',
                                                                                                           'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.loss_fn_3': ( 'methods/counternet.html#counternettrainingmodule.loss_fn_3',
                                                                                                           'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.pred_loss_fn': ( 'methods/counternet.html#counternettrainingmodule.pred_loss_fn',
                                                                                                              'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.predict': ( 'methods/counternet.html#counternettrainingmodule.predict',
                                                                                                         'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.training_step': ( 'methods/counternet.html#counternettrainingmodule.training_step',
                                                                                                               'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModule.validation_step': ( 'methods/counternet.html#counternettrainingmodule.validation_step',
                                                                                                                 'relax/methods/counternet.py'),
                                          'relax.methods.counternet.CounterNetTrainingModuleConfigs': ( 'methods/counternet.html#counternettrainingmoduleconfigs',
                                                                                                        'relax/methods/counternet.py'),
                                          'relax.methods.counternet.partition_trainable_params': ( 'methods/counternet.html#partition_trainable_params',
                                                                                                   'relax/methods/counternet.py'),
                                          'relax.methods.counternet.project_immutable_features': ( 'methods/counternet.html#project_immutable_features',
                                                                                                   'relax/methods/counternet.py')},
            'relax.methods.diverse': { 'relax.methods.diverse.DiverseCF': ('methods/diverse.html#diversecf', 'relax/methods/diverse.py'),
                                       'relax.methods.diverse.DiverseCF.__init__': ( 'methods/diverse.html#diversecf.__init__',
                                                                                     'relax/methods/diverse.py'),
                                       'relax.methods.diverse.DiverseCF.generate_cf': ( 'methods/diverse.html#diversecf.generate_cf',
                                                                                        'relax/methods/diverse.py'),
                                       'relax.methods.diverse.DiverseCF.generate_cfs': ( 'methods/diverse.html#diversecf.generate_cfs',
                                                                                         'relax/methods/diverse.py'),
                                       'relax.methods.diverse.DiverseCFConfig': ( 'methods/diverse.html#diversecfconfig',
                                                                                  'relax/methods/diverse.py'),
                                       'relax.methods.diverse.DiverseCFConfig.keys': ( 'methods/diverse.html#diversecfconfig.keys',
                                                                                       'relax/methods/diverse.py'),
                                       'relax.methods.diverse._compute_regularization_loss': ( 'methods/diverse.html#_compute_regularization_loss',
                                                                                               'relax/methods/diverse.py'),
                                       'relax.methods.diverse._diverse_cf': ( 'methods/diverse.html#_diverse_cf',
                                                                              'relax/methods/diverse.py'),
                                       'relax.methods.diverse.dpp_style': ('methods/diverse.html#dpp_style', 'relax/methods/diverse.py'),
                                       'relax.methods.diverse.hinge_loss': ('methods/diverse.html#hinge_loss', 'relax/methods/diverse.py'),
                                       'relax.methods.diverse.l1_mean': ('methods/diverse.html#l1_mean', 'relax/methods/diverse.py')},
            'relax.methods.proto': { 'relax.methods.proto.AE': ('methods/prototype.html#ae', 'relax/methods/proto.py'),
                                     'relax.methods.proto.AE.__call__': ('methods/prototype.html#ae.__call__', 'relax/methods/proto.py'),
                                     'relax.methods.proto.AE.__init__': ('methods/prototype.html#ae.__init__', 'relax/methods/proto.py'),
                                     'relax.methods.proto.AEConfigs': ('methods/prototype.html#aeconfigs', 'relax/methods/proto.py'),
                                     'relax.methods.proto.AETrainingModule': ( 'methods/prototype.html#aetrainingmodule',
                                                                               'relax/methods/proto.py'),
                                     'relax.methods.proto.AETrainingModule.__init__': ( 'methods/prototype.html#aetrainingmodule.__init__',
                                                                                        'relax/methods/proto.py'),
                                     'relax.methods.proto.AETrainingModule._training_step': ( 'methods/prototype.html#aetrainingmodule._training_step',
                                                                                              'relax/methods/proto.py'),
                                     'relax.methods.proto.AETrainingModule.encode': ( 'methods/prototype.html#aetrainingmodule.encode',
                                                                                      'relax/methods/proto.py'),
                                     'relax.methods.proto.AETrainingModule.forward': ( 'methods/prototype.html#aetrainingmodule.forward',
                                                                                       'relax/methods/proto.py'),
                                     'relax.methods.proto.AETrainingModule.init_net_opt': ( 'methods/prototype.html#aetrainingmodule.init_net_opt',
                                                                                            'relax/methods/proto.py'),
                                     'relax.methods.proto.AETrainingModule.loss_fn': ( 'methods/prototype.html#aetrainingmodule.loss_fn',
                                                                                       'relax/methods/proto.py'),
                                     'relax.methods.proto.AETrainingModule.training_step': ( 'methods/prototype.html#aetrainingmodule.training_step',
                                                                                             'relax/methods/proto.py'),
                                     'relax.methods.proto.AETrainingModule.validation_step': ( 'methods/prototype.html#aetrainingmodule.validation_step',
                                                                                               'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCF': ('methods/prototype.html#protocf', 'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCF.__init__': ( 'methods/prototype.html#protocf.__init__',
                                                                               'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCF._is_module_trained': ( 'methods/prototype.html#protocf._is_module_trained',
                                                                                         'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCF.generate_cf': ( 'methods/prototype.html#protocf.generate_cf',
                                                                                  'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCF.generate_cfs': ( 'methods/prototype.html#protocf.generate_cfs',
                                                                                   'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCF.train': ( 'methods/prototype.html#protocf.train',
                                                                            'relax/methods/proto.py'),
                                     'relax.methods.proto.ProtoCFConfig': ( 'methods/prototype.html#protocfconfig',
                                                                            'relax/methods/proto.py'),
                                     'relax.methods.proto._proto_cf': ('methods/prototype.html#_proto_cf', 'relax/methods/proto.py')},
            'relax.methods.sphere': { 'relax.methods.sphere.GSConfig': ('methods/sphere.html#gsconfig', 'relax/methods/sphere.py'),
                                      'relax.methods.sphere.GrowingSphere': ( 'methods/sphere.html#growingsphere',
                                                                              'relax/methods/sphere.py'),
                                      'relax.methods.sphere.GrowingSphere.__init__': ( 'methods/sphere.html#growingsphere.__init__',
                                                                                       'relax/methods/sphere.py'),
                                      'relax.methods.sphere.GrowingSphere.generate_cf': ( 'methods/sphere.html#growingsphere.generate_cf',
                                                                                          'relax/methods/sphere.py'),
                                      'relax.methods.sphere.GrowingSphere.generate_cfs': ( 'methods/sphere.html#growingsphere.generate_cfs',
                                                                                           'relax/methods/sphere.py'),
                                      'relax.methods.sphere._growing_spheres': ( 'methods/sphere.html#_growing_spheres',
                                                                                 'relax/methods/sphere.py'),
                                      'relax.methods.sphere.apply_immutable': ( 'methods/sphere.html#apply_immutable',
                                                                                'relax/methods/sphere.py'),
                                      'relax.methods.sphere.cat_sample': ('methods/sphere.html#cat_sample', 'relax/methods/sphere.py'),
                                      'relax.methods.sphere.hyper_sphere_coordindates': ( 'methods/sphere.html#hyper_sphere_coordindates',
                                                                                          'relax/methods/sphere.py')},
            'relax.methods.vaecf': { 'relax.methods.vaecf.Decoder': ('methods/vaecf.html#decoder', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.Decoder.__call__': ( 'methods/vaecf.html#decoder.__call__',
                                                                               'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.Decoder.__init__': ( 'methods/vaecf.html#decoder.__init__',
                                                                               'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.Encoder': ('methods/vaecf.html#encoder', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.Encoder.__call__': ( 'methods/vaecf.html#encoder.__call__',
                                                                               'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.Encoder.__init__': ( 'methods/vaecf.html#encoder.__init__',
                                                                               'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECF': ('methods/vaecf.html#vaecf', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECF.__init__': ('methods/vaecf.html#vaecf.__init__', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECF._is_module_trained': ( 'methods/vaecf.html#vaecf._is_module_trained',
                                                                                       'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECF.generate_cf': ( 'methods/vaecf.html#vaecf.generate_cf',
                                                                                'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECF.generate_cfs': ( 'methods/vaecf.html#vaecf.generate_cfs',
                                                                                 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECF.train': ('methods/vaecf.html#vaecf.train', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFConfigs': ('methods/vaecf.html#vaecfconfigs', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule': ('methods/vaecf.html#vaecfmodule', 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule.__init__': ( 'methods/vaecf.html#vaecfmodule.__init__',
                                                                                   'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule._training_step': ( 'methods/vaecf.html#vaecfmodule._training_step',
                                                                                         'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule.compute_loss': ( 'methods/vaecf.html#vaecfmodule.compute_loss',
                                                                                       'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule.decode': ( 'methods/vaecf.html#vaecfmodule.decode',
                                                                                 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule.encode': ( 'methods/vaecf.html#vaecfmodule.encode',
                                                                                 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule.init_net_opt': ( 'methods/vaecf.html#vaecfmodule.init_net_opt',
                                                                                       'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule.sample': ( 'methods/vaecf.html#vaecfmodule.sample',
                                                                                 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule.sample_latent_code': ( 'methods/vaecf.html#vaecfmodule.sample_latent_code',
                                                                                             'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule.sample_step': ( 'methods/vaecf.html#vaecfmodule.sample_step',
                                                                                      'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule.training_step': ( 'methods/vaecf.html#vaecfmodule.training_step',
                                                                                        'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModule.validation_step': ( 'methods/vaecf.html#vaecfmodule.validation_step',
                                                                                          'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.VAECFModuleConfigs': ( 'methods/vaecf.html#vaecfmoduleconfigs',
                                                                                 'relax/methods/vaecf.py'),
                                     'relax.methods.vaecf.hindge_embedding_loss': ( 'methods/vaecf.html#hindge_embedding_loss',
                                                                                    'relax/methods/vaecf.py')},
            'relax.methods.vanilla': { 'relax.methods.vanilla.VanillaCF': ('methods/vanilla.html#vanillacf', 'relax/methods/vanilla.py'),
                                       'relax.methods.vanilla.VanillaCF.__init__': ( 'methods/vanilla.html#vanillacf.__init__',
                                                                                     'relax/methods/vanilla.py'),
                                       'relax.methods.vanilla.VanillaCF.generate_cf': ( 'methods/vanilla.html#vanillacf.generate_cf',
                                                                                        'relax/methods/vanilla.py'),
                                       'relax.methods.vanilla.VanillaCF.generate_cfs': ( 'methods/vanilla.html#vanillacf.generate_cfs',
                                                                                         'relax/methods/vanilla.py'),
                                       'relax.methods.vanilla.VanillaCFConfig': ( 'methods/vanilla.html#vanillacfconfig',
                                                                                  'relax/methods/vanilla.py'),
                                       'relax.methods.vanilla._vanilla_cf': ( 'methods/vanilla.html#_vanilla_cf',
                                                                              'relax/methods/vanilla.py')},
            'relax.module': { 'relax.module.BaseNetwork': ('training_module.html#basenetwork', 'relax/module.py'),
                              'relax.module.BaseNetwork.__call__': ('training_module.html#basenetwork.__call__', 'relax/module.py'),
                              'relax.module.BaseTrainingModule': ('training_module.html#basetrainingmodule', 'relax/module.py'),
                              'relax.module.BaseTrainingModule.init_logger': ( 'training_module.html#basetrainingmodule.init_logger',
                                                                               'relax/module.py'),
                              'relax.module.BaseTrainingModule.init_net_opt': ( 'training_module.html#basetrainingmodule.init_net_opt',
                                                                                'relax/module.py'),
                              'relax.module.BaseTrainingModule.log': ('training_module.html#basetrainingmodule.log', 'relax/module.py'),
                              'relax.module.BaseTrainingModule.log_dict': ( 'training_module.html#basetrainingmodule.log_dict',
                                                                            'relax/module.py'),
                              'relax.module.BaseTrainingModule.save_hyperparameters': ( 'training_module.html#basetrainingmodule.save_hyperparameters',
                                                                                        'relax/module.py'),
                              'relax.module.BaseTrainingModule.training_step': ( 'training_module.html#basetrainingmodule.training_step',
                                                                                 'relax/module.py'),
                              'relax.module.BaseTrainingModule.validation_step': ( 'training_module.html#basetrainingmodule.validation_step',
                                                                                   'relax/module.py'),
                              'relax.module.DenseBlock': ('training_module.html#denseblock', 'relax/module.py'),
                              'relax.module.DenseBlock.__call__': ('training_module.html#denseblock.__call__', 'relax/module.py'),
                              'relax.module.DenseBlock.__init__': ('training_module.html#denseblock.__init__', 'relax/module.py'),
                              'relax.module.MLP': ('training_module.html#mlp', 'relax/module.py'),
                              'relax.module.MLP.__call__': ('training_module.html#mlp.__call__', 'relax/module.py'),
                              'relax.module.MLP.__init__': ('training_module.html#mlp.__init__', 'relax/module.py'),
                              'relax.module.PredictiveModel': ('training_module.html#predictivemodel', 'relax/module.py'),
                              'relax.module.PredictiveModel.__call__': ('training_module.html#predictivemodel.__call__', 'relax/module.py'),
                              'relax.module.PredictiveModel.__init__': ('training_module.html#predictivemodel.__init__', 'relax/module.py'),
                              'relax.module.PredictiveModelConfigs': ('training_module.html#predictivemodelconfigs', 'relax/module.py'),
                              'relax.module.PredictiveTrainingModule': ('training_module.html#predictivetrainingmodule', 'relax/module.py'),
                              'relax.module.PredictiveTrainingModule.__init__': ( 'training_module.html#predictivetrainingmodule.__init__',
                                                                                  'relax/module.py'),
                              'relax.module.PredictiveTrainingModule._training_step': ( 'training_module.html#predictivetrainingmodule._training_step',
                                                                                        'relax/module.py'),
                              'relax.module.PredictiveTrainingModule.forward': ( 'training_module.html#predictivetrainingmodule.forward',
                                                                                 'relax/module.py'),
                              'relax.module.PredictiveTrainingModule.init_net_opt': ( 'training_module.html#predictivetrainingmodule.init_net_opt',
                                                                                      'relax/module.py'),
                              'relax.module.PredictiveTrainingModule.loss_fn': ( 'training_module.html#predictivetrainingmodule.loss_fn',
                                                                                 'relax/module.py'),
                              'relax.module.PredictiveTrainingModule.pred_fn': ( 'training_module.html#predictivetrainingmodule.pred_fn',
                                                                                 'relax/module.py'),
                              'relax.module.PredictiveTrainingModule.training_step': ( 'training_module.html#predictivetrainingmodule.training_step',
                                                                                       'relax/module.py'),
                              'relax.module.PredictiveTrainingModule.validation_step': ( 'training_module.html#predictivetrainingmodule.validation_step',
                                                                                         'relax/module.py'),
                              'relax.module.PredictiveTrainingModuleConfigs': ( 'training_module.html#predictivetrainingmoduleconfigs',
                                                                                'relax/module.py')},
            'relax.plots': { 'relax.plots._barplot': ('plotting.html#_barplot', 'relax/plots.py'),
                             'relax.plots._diff': ('plotting.html#_diff', 'relax/plots.py'),
                             'relax.plots._process_exp': ('plotting.html#_process_exp', 'relax/plots.py'),
                             'relax.plots._stripplot': ('plotting.html#_stripplot', 'relax/plots.py'),
                             'relax.plots._swarmplot': ('plotting.html#_swarmplot', 'relax/plots.py'),
                             'relax.plots.individual_plot': ('plotting.html#individual_plot', 'relax/plots.py'),
                             'relax.plots.summary_plot': ('plotting.html#summary_plot', 'relax/plots.py')},
            'relax.trainer': { 'relax.trainer.TrainingConfigs': ('learning.html#trainingconfigs', 'relax/trainer.py'),
                               'relax.trainer.TrainingConfigs.PRNGSequence': ( 'learning.html#trainingconfigs.prngsequence',
                                                                               'relax/trainer.py'),
                               'relax.trainer.train_model': ('learning.html#train_model', 'relax/trainer.py'),
                               'relax.trainer.train_model_with_states': ('learning.html#train_model_with_states', 'relax/trainer.py')},
            'relax.utils': { 'relax.utils.Config': ('utils.html#config', 'relax/utils.py'),
                             'relax.utils.Config.default': ('utils.html#config.default', 'relax/utils.py'),
                             'relax.utils.ParserMarkdownRenderer': ('utils.html#parsermarkdownrenderer', 'relax/utils.py'),
                             'relax.utils.ParserMarkdownRenderer.__init__': ( 'utils.html#parsermarkdownrenderer.__init__',
                                                                              'relax/utils.py'),
                             'relax.utils._docment_parser': ('utils.html#_docment_parser', 'relax/utils.py'),
                             'relax.utils._reshape_x': ('utils.html#_reshape_x', 'relax/utils.py'),
                             'relax.utils.accuracy': ('utils.html#accuracy', 'relax/utils.py'),
                             'relax.utils.auto_reshaping': ('utils.html#auto_reshaping', 'relax/utils.py'),
                             'relax.utils.binary_cross_entropy': ('utils.html#binary_cross_entropy', 'relax/utils.py'),
                             'relax.utils.cat_normalize': ('utils.html#cat_normalize', 'relax/utils.py'),
                             'relax.utils.check_cat_info': ('utils.html#check_cat_info', 'relax/utils.py'),
                             'relax.utils.dist': ('utils.html#dist', 'relax/utils.py'),
                             'relax.utils.get_config': ('utils.html#get_config', 'relax/utils.py'),
                             'relax.utils.grad_update': ('utils.html#grad_update', 'relax/utils.py'),
                             'relax.utils.init_net_opt': ('utils.html#init_net_opt', 'relax/utils.py'),
                             'relax.utils.load_json': ('utils.html#load_json', 'relax/utils.py'),
                             'relax.utils.make_hk_module': ('utils.html#make_hk_module', 'relax/utils.py'),
                             'relax.utils.make_model': ('utils.html#make_model', 'relax/utils.py'),
                             'relax.utils.proximity': ('utils.html#proximity', 'relax/utils.py'),
                             'relax.utils.show_doc': ('utils.html#show_doc', 'relax/utils.py'),
                             'relax.utils.sigmoid': ('utils.html#sigmoid', 'relax/utils.py'),
                             'relax.utils.validate_configs': ('utils.html#validate_configs', 'relax/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/01_data.loader.ipynb.

# %% ../../nbs/01_data.loader.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from ..utils import get_config

# %% auto 0
__all__ = ['Dataset', 'ArrayDataset', 'BaseDataLoader', 'JaxDataloader', 'TorchDataloader', 'DataloaderBackends', 'DataLoader']

# %% ../../nbs/01_data.loader.ipynb 4
try:
    import torch.utils.data as torch_data
except ModuleNotFoundError:
    torch_data = None

# %% ../../nbs/01_data.loader.ipynb 6
class Dataset(ABC):
    """A pytorch-like abstract Dataset class."""

    def __getitem__(self, index):
        raise NotImplementedError

# %% ../../nbs/01_data.loader.ipynb 7
class ArrayDataset(Dataset):
    """Dataset wrapping tensors."""

    def __init__(
        self, 
        *arrays: jnp.DeviceArray # Numpy array with same first dimension
    ):
        assert all(arrays[0].shape[0] == arr.shape[0] for arr in arrays), \
            "All arrays must have the same dimension."
        self.arrays = arrays

    def __len__(self):
        return self.arrays[0].shape[0]

    def __getitem__(self, index):
        return tuple(arr[index] for arr in self.arrays)

# %% ../../nbs/01_data.loader.ipynb 12
class BaseDataLoader(ABC):
    """Dataloader Interface"""
    def __init__(
        self, 
        dataset,
        backend: str,
        batch_size: int = 1,    # Batch size
        shuffle: bool = False,  # If true, dataloader shuffles before sampling each batch
        drop_last: bool = False, # Drop last batches or not
        **kwargs # Aux arguments
    ):
        pass 
    
    def __len__(self):
        raise NotImplementedError
    
    def __next__(self):
        raise NotImplementedError
    
    def __iter__(self):
        raise NotImplementedError

# %% ../../nbs/01_data.loader.ipynb 14
class JaxDataloader(BaseDataLoader):
    """Dataloder in vanilla Jax"""

    def __init__(
        self, 
        dataset: Dataset,
        backend: str = 'jax', # Position argument
        batch_size: int = 1,  # Batch size
        shuffle: bool = False,  # If true, dataloader shuffles before sampling each batch
        drop_last: bool = False, # Drop last batches or not
        **kwargs
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.drop_last = drop_last

        self.keys = hk.PRNGSequence(get_config().global_seed)
        self.data_len = len(dataset)  # Length of the dataset
        self.indices = jnp.arange(self.data_len) # available indices in the dataset
        self.pose = 0  # record the current position in the dataset
        self._shuffle()

    def _shuffle(self):
        if self.shuffle:
            self.indices = jax.random.permutation(next(self.keys), self.indices)
        
    def _stop_iteration(self):
        self.pose = 0
        self._shuffle()
        raise StopIteration

    def __len__(self):
        if self.drop_last:
            batches = len(self.dataset) // self.batch_size  # get the floor of division
        else:
            batches = -(len(self.dataset) // -self.batch_size)  # get the ceil of division
        return batches

    def __next__(self):
        if self.pose + self.batch_size <= self.data_len:
            batch_indices = self.indices[self.pose: self.pose + self.batch_size]
            batch_data = self.dataset[batch_indices, ...]
            self.pose += self.batch_size
            return batch_data
        elif self.pose < self.data_len and not self.drop_last:
            batch_indices = self.indices[self.pose:]
            batch_data = self.dataset[batch_indices, ...]
            self.pose += self.batch_size
            return batch_data
        else:
            self._stop_iteration()

    def __iter__(self):
        return self

# %% ../../nbs/01_data.loader.ipynb 16
# copy from https://jax.readthedocs.io/en/latest/notebooks/Neural_Network_and_Data_Loading.html#data-loading-with-pytorch
def _numpy_collate(batch):
    if isinstance(batch[0], np.ndarray):
        return np.stack(batch)
    elif isinstance(batch[0], (tuple, list)):
        transposed = zip(*batch)
        return [_numpy_collate(samples) for samples in transposed]
    else:
        return np.array(batch)

def _convert_dataset_pytorch(dataset: Dataset):
    class _TorchDataset(torch_data.Dataset):
        def __init__(self, dataset: Dataset): self.dataset = dataset
        def __len__(self): return len(self.dataset)
        def __getitem__(self, idx): return self.dataset[idx]
    
    return _TorchDataset(dataset)

# %% ../../nbs/01_data.loader.ipynb 17
class TorchDataloader(BaseDataLoader):
    """Use `Pytorch` to load batches. It requires [pytorch](https://pytorch.org/get-started/) to be installed."""
    
    def __init__(
        self, 
        dataset: Dataset,
        backend: str = 'pytorch', # positional argument
        batch_size: int = 1,  # batch size
        shuffle: bool = False,  # if true, dataloader shuffles before sampling each batch
        num_workers: int = 0, # number of workers
        drop_last: bool = False, # drop last batch or not
        **kwargs
    ):
        if torch_data is None:
            raise ModuleNotFoundError("`pytorch` library needs to be installed. Try `pip install torch`."
            "Please refer to pytorch documentation for details: https://pytorch.org/get-started/.")
        
        dataset = _convert_dataset_pytorch(dataset)
        self.dataloader = torch_data.DataLoader(
            dataset, 
            batch_size=batch_size, 
            shuffle=shuffle, 
            num_workers=num_workers, 
            drop_last=drop_last,
            collate_fn=_numpy_collate,
            **kwargs
        ) 

    def __len__(self):
        return len(self.dataloader)

    def __next__(self):
        return next(self.dataloader)

    def __iter__(self):
        return self.dataloader.__iter__()

# %% ../../nbs/01_data.loader.ipynb 20
@dataclass(frozen=True)
class DataloaderBackends:
    jax: BaseDataLoader = JaxDataloader
    pytorch: BaseDataLoader = TorchDataloader
    tensorflow: BaseDataLoader = None
    merlin: BaseDataLoader = None

    __all__ = dict(
        jax=jax, pytorch=pytorch, tensorflow=tensorflow, merlin=merlin
    )

    def __getitem__(self, key):
        return self.__all__[key]

    @classmethod
    def supported(cls) -> List[str]:
        return [
            backend for backend, dl_cls in cls.__all__.items() if dl_cls is not None
        ]

# %% ../../nbs/01_data.loader.ipynb 21
def _dispatch_dataloader(
    backend: str # dataloader backend
) -> BaseDataLoader:
    """Return Dataloader class based on given `backend`"""

    backends = DataloaderBackends()
    if not backend in DataloaderBackends.supported():
        raise ValueError(f"backend=`{backend}` is either an invalid backend or not supported yet. "
            f"Should be one of {backends.supported}.")
    
    dl_cls = backends[backend]
    return dl_cls

# %% ../../nbs/01_data.loader.ipynb 24
class DataLoader(BaseDataLoader):
    """Main Dataloader class to load Numpy data batches"""
    def __init__(
        self,
        dataset: Dataset,
        backend: str, # Dataloader backend; Currently supports `['jax', 'pytorch']`
        batch_size: int = 1,  # batch size
        shuffle: bool = False,  # if true, dataloader shuffles before sampling each batch
        num_workers: int = 0, # number of workers
        drop_last: bool = False, # drop last batches or not
        **kwargs
    ):
        self.__class__ = _dispatch_dataloader(backend)
        self.__init__(
            dataset=dataset, 
            backend=backend, 
            batch_size=batch_size, 
            shuffle=shuffle, 
            num_workers=num_workers,
            drop_last=drop_last,
            **kwargs
        )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/05_sphere.ipynb.

# %% ../../nbs/methods/05_sphere.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from .base import BaseCFModule
from ..utils import *

# %% auto 0
__all__ = ['hyper_sphere_coordindates', 'cat_sample', 'apply_immutable', 'GSConfig', 'GrowingSphere']

# %% ../../nbs/methods/05_sphere.ipynb 4
def hyper_sphere_coordindates(
    rng_key: jrand.PRNGKey, # Random number generator key
    x: Array, # Input instance with only continuous features. Shape: (1, n_features)
    n_samples: int, # Number of samples
    high: float, # Upper bound
    low: float, # Lower bound
    p_norm: int = 2 # Norm
):
    # Adapted from 
    # https://github.com/carla-recourse/CARLA/blob/24db00aa8616eb2faedea0d6edf6e307cee9d192/carla/recourse_methods/catalog/growing_spheres/library/gs_counterfactuals.py#L8
    key_1, key_2 = jrand.split(rng_key)
    delta = jrand.normal(key_1, shape=(n_samples, x.shape[-1]))
    dist = jrand.uniform(key_2, shape=(n_samples,)) * (high - low) + low
    norm_p = jnp.linalg.norm(delta, ord=p_norm, axis=1)
    d_norm = jnp.divide(dist, norm_p).reshape(-1, 1)  # rescale/normalize factor
    delta = jnp.multiply(delta, d_norm)
    candidates = x + delta

    return candidates

# %% ../../nbs/methods/05_sphere.ipynb 5
def cat_sample(
    rng_key: jrand.PRNGKey, # Random number generator key
    x: Array, # Input instance with only categorical features. Shape: (1, n_features)
    cat_arrays: List[List[str]],  # A list of a list of each categorical feature name
    n_samples: int,  # Number of samples to sample
): 
    def sample_categorical(rng_key: jrand.PRNGKey, col: np.ndarray):
        rng_key, subkey = jrand.split(rng_key)
        prob = jnp.ones(len(col)) / len(col)
        cat_sample = jax.nn.one_hot(
            jrand.categorical(rng_key, prob, shape=(n_samples,)), num_classes=len(col)
        )
        return rng_key, cat_sample
    
    candidates = []
    # We cannot use lax.scan here because cat_arrays is List[List[str]], not and can't ben an Array
    for col in cat_arrays:
        rng_key, cat_sample = sample_categorical(rng_key, col)
        candidates.append(cat_sample)
    candidates = jnp.concatenate(candidates, axis=1)
    return candidates

# %% ../../nbs/methods/05_sphere.ipynb 6
@auto_reshaping('x')
def _growing_spheres(
    rng_key: jrand.PRNGKey, # Random number generator key
    x: Array, # Input instance. Shape: (n_features)
    pred_fn: Callable, # Prediction function
    n_steps: int, # Number of steps
    n_samples: int,  # Number of samples to sample
    cat_idx: int, # Index of categorical features
    cat_arrays: List[List[str]],  # A list of a list of each categorical feature name
    step_size: float, # Step size
    p_norm: int, # Norm
    apply_fn: Callable # Apply immutable constraints
):
    @jit
    def cond_fn(state):
        candidate_cf, count, _ = state
        return (jnp.any(jnp.isinf(candidate_cf))) & (count < n_steps)
        # return (not candidate_cf) & (count < n_steps)
    
    @jit
    def body_fn(state):
        candidate_cf, count, rng_key = state
        rng_key, subkey_1, subkey_2 = jrand.split(rng_key, num=3)
        low, high = step_size * count, step_size * (count + 1)
        # Sample around x
        cont_candidates = hyper_sphere_coordindates(subkey_1, x[:, :cat_idx], n_samples, high, low, p_norm)
        cat_candidates = cat_sample(subkey_2, x[:, cat_idx:], cat_arrays, n_samples)
        candidates = jnp.concatenate([cont_candidates, cat_candidates], axis=1)
        # Apply immutable constraints
        candidates = apply_fn(x=x, cf=candidates)
        assert candidates.shape[1] == x.shape[1], f"candidates.shape = {candidates.shape}, x.shape = {x.shape}"

        # Calculate distance
        if p_norm == 1:
            dist = jnp.abs(candidates - x).sum(axis=1)
        elif p_norm == 2:
            dist = jnp.linalg.norm(candidates - x, ord=2, axis=1)
        else:
            raise ValueError("Only p_norm = 1 or 2 is supported")

        # Calculate counterfactual labels
        candidate_preds = pred_fn(candidates).round().reshape(-1)
        # print(candidate_preds != y_pred)
        indices = jnp.where(candidate_preds != y_pred, 1, 0).astype(bool)

        # candidates = candidates[indices]
        # candidates = jnp.where(indices.reshape(-1, 1), 
        candidates = jnp.where(indices.reshape(-1, 1), 
                               candidates, jnp.ones_like(candidates) * jnp.inf)
        # dist = dist[indices]
        dist = jnp.where(indices.reshape(-1, 1), dist, jnp.ones_like(dist) * jnp.inf)

        # if len(candidates) > 0:
        closest_idx = dist.argmin()
        candidate_cf = candidates[closest_idx].reshape(1, -1)

        # if jnp.any(jnp.logical_not(jnp.isinf(candidates))):
        #     # Find the closest counterfactual
        #     closest_idx = dist.argmin()
        #     candidate_cf = candidates[closest_idx].reshape(1, -1)

        return candidate_cf, count + 1, rng_key
    
    y_pred = pred_fn(x).round().reshape(-1)
    candidate_cf = jnp.ones_like(x) * jnp.inf
    count = 0
    state = (candidate_cf, count, rng_key)
    candidate_cf, _, _ = lax.while_loop(cond_fn, body_fn, state)
    # if `inf` is found, return the original input
    candidate_cf = jnp.where(jnp.isinf(candidate_cf), x, candidate_cf)
    return candidate_cf

# %% ../../nbs/methods/05_sphere.ipynb 7
def apply_immutable(x: Array, cf: Array, immutable_idx: List[int]):
    if immutable_idx is not None:
        cf = cf.at[:, immutable_idx].set(x[:, immutable_idx])
    return cf

# %% ../../nbs/methods/05_sphere.ipynb 8
class GSConfig(BaseParser):
    seed: int = 42
    n_steps: int = 100
    n_samples: int = 1000
    step_size: float = 0.05
    p_norm: int = 2
    

# %% ../../nbs/methods/05_sphere.ipynb 9
class GrowingSphere(BaseCFModule):
    name = "Growing Sphere"

    def __init__(
        self,
        configs: Dict | GSConfig = None
    ):
        if configs is None:
            configs = GSConfig()
        self.configs = validate_configs(configs, GSConfig)
        self.rng = jrand.PRNGKey(self.configs.seed)
    
    def generate_cf(
        self,
        x: Array,
        pred_fn: Callable,
    ):
        cat_idx = self.data_module.cat_idx
        apply_immutable_partial = partial(
            apply_immutable, immutable_idx=self.data_module._imutable_idx_list)
        cf = _growing_spheres(
            self.rng,
            x,
            pred_fn,
            self.configs.n_steps,
            self.configs.n_samples,
            cat_idx,
            self.data_module._cat_arrays,
            self.configs.step_size,
            self.configs.p_norm,
            apply_immutable_partial,
        )
        return cf
    
    def generate_cfs(
        self, 
        X: Array, 
        pred_fn: Callable = None
    ) -> jnp.ndarray:
        rng_keys = jrand.split(jrand.PRNGKey(self.configs.seed), num=X.shape[0])
        generate_cf_partial = jit(partial(self.generate_cf, pred_fn=pred_fn))
        cfs = jax.vmap(generate_cf_partial)(X, rng_keys)
        # cfs = generate_cf_partial(X[0], rng_keys[0])
        return cfs



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/07_plotting.ipynb.

# %% ../nbs/07_plotting.ipynb 3
from __future__ import annotations
from .import_essentials import *
import seaborn as sns
import matplotlib.pyplot as plt
from .evaluate import Explanation
from .data import TabularDataModule

# %% auto 0
__all__ = ['summary_plot', 'individual_plot']

# %% ../nbs/07_plotting.ipynb 4
def _process_exp(
    x: jnp.DeviceArray, # Dim: (N, k)
    cf: jnp.DeviceArray, # Dim: (N, k)
    pred_fn: Callable, # Predictive function
    return_valid: bool = False # Return only valid `x`, `cf`, `y_cf` or not
): # Return `x`, `cf`, `y_cf`
    """Process valid data."""
    
    assert x.shape == cf.shape
    assert len(x.shape) > 1

    y_cf = pred_fn(cf).round()
    
    if return_valid:
        y_x = pred_fn(x).round()
        idx = (y_x == y_cf).reshape(-1)
        x, cf, y_cf = x[idx], cf[idx], y_cf[idx]
    return x, cf, y_cf

# %% ../nbs/07_plotting.ipynb 6
def _diff(
    x: jnp.DeviceArray, # Dim: (N, k)
    cf: jnp.DeviceArray, # Dim: (N, k)
    pred_fn: Callable, # Predictive function
    datamodule: TabularDataModule, # Data module
    return_valid: bool = False, # Return only valid `x`, `cf`, `y_cf` or not
    return_outcome_col: bool = True # Return outcome column or not
) -> Tuple[pd.DataFrame, pd.DataFrame]: # Return continuous and categorical differences
    """Calculate differences between `x` and `cf`"""
    
    x, cf, y_cf = _process_exp(x, cf, pred_fn, return_valid)
    # wrap to numpy.array
    x, cf, y_cf = map(np.asarray, (x, cf, y_cf))
    x_inv, cf_inv = map(datamodule.inverse_transform, [x, cf], [y_cf, y_cf])
    cat_idx = datamodule.cat_idx
    cont_diff = pd.DataFrame(
        data=cf[:, :cat_idx] - x[:, :cat_idx],
        columns=datamodule._configs.continous_cols
    ) 
    cat_diff = (~cf_inv.iloc[:, cat_idx:-1].eq(x_inv.iloc[:, cat_idx:-1])).astype(float)
    if return_outcome_col:
        return pd.concat([cont_diff, cf_inv.iloc[:, -1:]], axis=1), \
            pd.concat([cat_diff, cf_inv.iloc[:, -1:]], axis=1)
    else:
        return cont_diff, cat_diff


# %% ../nbs/07_plotting.ipynb 7
def _swarmplot(
    cont_diff: pd.DataFrame, ax
):
    target_col = cont_diff.columns[-1]
    cont_diff = cont_diff.melt(
        id_vars=[target_col], value_name='Difference', var_name='Features'
    )
    sns.stripplot(
        y="Features",
        x="Difference",
        data=cont_diff,
        hue=target_col,
        alpha=0.5,
        size=5,
        jitter=0.3,
        ax=ax
    ).set(title="Summary Plot for Continuous Features")
    return ax

# %% ../nbs/07_plotting.ipynb 8
def _stripplot(
    cat_diff: pd.DataFrame, ax
):
    target_col = cat_diff.columns[-1]
    cat_diff = cat_diff.melt(
        id_vars=[target_col], value_name='Difference', var_name='Features'
    )
    # add noise for clear display 
    jitter = 0.3
    delta = np.random.uniform(-jitter/2, jitter/2, len(cat_diff['Difference']))
    cat_diff['Difference'] = cat_diff['Difference'].values + delta
    sns.stripplot(
        y="Features",
        x="Difference",
        data=cat_diff,
        hue=target_col,
        alpha=0.5,
        size=5,
        jitter=0.3,
        ax=ax
    ).set(title="Summary Plot for Categorical Features")
    return ax

# %% ../nbs/07_plotting.ipynb 9
def summary_plot(
    exp: Explanation, # Explanations to visualize
    sample_frac: float = 1.0, # Sample part of data for visualization
    only_valid: bool = False, # Use only valid data
    figsize=(15, 7) # Figure size
) -> plt.Figure:
    """Globally visualize generated explanations."""

    cont, cat = _diff(
        exp.X, exp.cfs, exp.pred_fn, exp.data_module, only_valid
    )
    fig, (ax1, ax2) = plt.subplots(1, 2)
    _swarmplot(cont.sample(frac=sample_frac), ax1)
    _stripplot(cat.sample(frac=sample_frac), ax2)

    # Resize to create overlap
    fig.set_size_inches(figsize[0], figsize[1])
    plt.subplots_adjust(wspace=.5)

    return fig

# %% ../nbs/07_plotting.ipynb 10
def _barplot(
    diff: pd.Series, ax
):
    sns.barplot(
        x=diff.values, y=diff.index, ax=ax
    ).set(title="Counterfactual Explanations for An Individual Sample")

# %% ../nbs/07_plotting.ipynb 11
def individual_plot(
    exp: Explanation, # Explanations to visualize
    idx: int, # Index for visualization
    figsize=(15, 7) # Figure plot
) -> plt.Figure:
    """Locally visualize individual explanations."""

    cont, cat = _diff(
        exp.X[idx, :].reshape(1, -1), exp.cfs[idx, :].reshape(1, -1), 
        exp.pred_fn, exp.data_module, return_outcome_col=False
    )
    diff = pd.concat([cont, cat], axis=1)

    # plot individual differences
    fig, ax = plt.subplots()
    _barplot(diff.iloc[0], ax)
    fig.set_size_inches(figsize[0], figsize[1])
    return fig



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_utils.ipynb.

# %% ../nbs/00_utils.ipynb 3
from __future__ import annotations
from .import_essentials import *
import nbdev
from fastcore.basics import AttrDict
from nbdev.showdoc import BasicMarkdownRenderer
from inspect import isclass
from fastcore.test import *
from jax.core import InconclusiveDimensionOperation

# %% auto 0
__all__ = ['validate_configs', 'cat_normalize', 'auto_reshaping', 'make_model', 'make_hk_module', 'init_net_opt', 'grad_update',
           'check_cat_info', 'load_json', 'binary_cross_entropy', 'sigmoid', 'accuracy', 'dist', 'proximity',
           'get_config']

# %% ../nbs/00_utils.ipynb 5
def validate_configs(
    configs: dict | BaseParser,  # A configuration of the model/dataset.
    config_cls: BaseParser,  # The desired configuration class.
) -> BaseParser:
    """return a valid configuration object."""

    assert isclass(config_cls), f"`config_cls` should be a class."
    assert issubclass(config_cls, BaseParser), \
        f"{config_cls} should be a subclass of `BaseParser`."
    
    if isinstance(configs, dict):
        configs = config_cls(**configs)
    if not isinstance(configs, config_cls):
        raise TypeError(
            f"configs should be either a `dict` or an instance of {config_cls.__name__}.")
    return configs

# %% ../nbs/00_utils.ipynb 14
def _docment_parser(parser: BaseParser):
    p = parser.schema()['properties']
    anno = parser.__annotations__
    d = { 
        k: {
            'anno': anno[k],
            'default': v['default'] if 'default' in v else inspect._empty,
            'docment': v['description'] if 'description' in v else inspect._empty,
        } for k, v in p.items()
    }

    d = AttrDict(d)
    return d


# %% ../nbs/00_utils.ipynb 15
class ParserMarkdownRenderer(BasicMarkdownRenderer):
        def __init__(self, sym, name: str | None = None, title_level: int = 3):
            super().__init__(sym, name, title_level)
            self.dm.dm = _docment_parser(sym)

# %% ../nbs/00_utils.ipynb 16
def show_doc(
    sym # Symbol to document
):
    """Same functionality as [nbdev.show_doc](configurator), 
    but provide additional support for `BaseParser`."""
    if inspect.isclass(sym) and issubclass(sym, BaseParser):
        return nbdev.show_doc(sym, ParserMarkdownRenderer)
    else:
        return nbdev.show_doc(sym)

# %% ../nbs/00_utils.ipynb 19
def cat_normalize(
    cf: jnp.ndarray,  # Unnormalized counterfactual explanations `[n_samples, n_features]`
    cat_arrays: List[List[str]],  # A list of a list of each categorical feature name
    cat_idx: int,  # Index that starts categorical features
    hard: bool = False,  # If `True`, return one-hot vectors; If `False`, return probability normalized via softmax
) -> jnp.ndarray:
    """Ensure generated counterfactual explanations to respect one-hot encoding constraints."""
    cf_cont = cf[:, :cat_idx]
    normalized_cf = [cf_cont]

    for col in cat_arrays:
        cat_end_idx = cat_idx + len(col)
        _cf_cat = cf[:, cat_idx:cat_end_idx]

        cf_cat = lax.cond(
            hard,
            true_fun=lambda x: jax.nn.one_hot(jnp.argmax(x, axis=-1), len(col)),
            false_fun=lambda x: jax.nn.softmax(x, axis=-1),
            operand=_cf_cat,
        )

        cat_idx = cat_end_idx
        normalized_cf.append(cf_cat)
    return jnp.concatenate(normalized_cf, axis=-1)


# %% ../nbs/00_utils.ipynb 33
def _reshape_x(x: Array):
    x_size = x.shape
    if len(x_size) > 1 and x_size[0] != 1:
        raise ValueError(
            f"""Invalid Input Shape: Require `x.shape` = (1, k) or (k, ),
but got `x.shape` = {x.shape}. This method expects a single input instance."""
        )
    if len(x_size) == 1:
        x = x.reshape(1, -1)
    return x, x_size

# %% ../nbs/00_utils.ipynb 34
def auto_reshaping(reshape_argname: str):
    """
    Decorator to automatically reshape function's input into (1, k), 
    and out to input's shape.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            kwargs = inspect.getcallargs(func, *args, **kwargs)
            if reshape_argname in kwargs:
                reshaped_x, x_shape = _reshape_x(kwargs[reshape_argname])
                kwargs[reshape_argname] = reshaped_x
            else:
                raise ValueError(
                    f"Invalid argument name: `{reshape_argname}` is not a valid argument name.")
            cf = func(**kwargs)
            if not isinstance(cf, Array): 
                raise ValueError(
                    f"Invalid return type: must be a `jax.Array`, but got `{type(cf).__name__}`.")
            try: 
                cf = cf.reshape(x_shape)
            except InconclusiveDimensionOperation:
                raise ValueError(
                    f"Invalid return shape: Require `cf.shape` = {cf.shape} "
                    f"is not compatible with `x.shape` = {x_shape}.")
            return cf

        return wrapper
    return decorator

# %% ../nbs/00_utils.ipynb 39
def make_model(
    m_configs: Dict[str, Any], model: hk.Module  # model configs
) -> hk.Transformed:
    # example:
    # net = make_model(PredictiveModel)
    # params = net.init(...)
    def model_fn(x, is_training: bool = True):
        return model(m_configs)(x, is_training)

    return hk.transform(model_fn)


# %% ../nbs/00_utils.ipynb 40
def make_hk_module(
    module: hk.Module, # haiku module 
    *args, # haiku module arguments
    **kargs, # haiku module arguments
) -> hk.Transformed:

    def model_fn(x, is_training: bool = True):
        return module(*args, **kargs)(x, is_training)
    
    return hk.transform(model_fn)


# %% ../nbs/00_utils.ipynb 41
def init_net_opt(
    net: hk.Transformed,
    opt: optax.GradientTransformation,
    X: jnp.DeviceArray,
    key: random.PRNGKey,
) -> Tuple[hk.Params, optax.OptState]:
    X = device_put(X)
    params = net.init(key, X, is_training=True)
    opt_state = opt.init(params)
    return params, opt_state


# %% ../nbs/00_utils.ipynb 42
def grad_update(
    grads: Dict[str, jnp.ndarray],
    params: hk.Params,
    opt_state: optax.OptState,
    opt: optax.GradientTransformation,
) -> Tuple[hk.Params, optax.OptState]:
    updates, opt_state = opt.update(grads, opt_state)
    upt_params = optax.apply_updates(params, updates)
    return upt_params, opt_state


# %% ../nbs/00_utils.ipynb 43
def check_cat_info(method):
    def inner(cf_module, *args, **kwargs):
        warning_msg = f"""This CFExplanationModule might not be updated with categorical information.
You should try `{cf_module.name}.update_cat_info(dm)` before generating cfs.
        """
        if cf_module.cat_idx == 0 and cf_module.cat_arrays == []:
            warnings.warn(warning_msg, RuntimeWarning)
        return method(cf_module, *args, **kwargs)

    return inner


# %% ../nbs/00_utils.ipynb 45
def load_json(f_name: str) -> Dict[str, Any]:  # file name
    with open(f_name) as f:
        return json.load(f)


# %% ../nbs/00_utils.ipynb 47
def binary_cross_entropy(
    preds: jnp.DeviceArray, # The predicted values
    labels: jnp.DeviceArray # The ground-truth labels
) -> jnp.DeviceArray: # Loss value
    """Per-sample binary cross-entropy loss function."""

    # Clip the predictions to avoid NaNs in the log
    preds = jnp.clip(preds, 1e-7, 1 - 1e-7)

    # Compute the binary cross-entropy
    loss = -labels * jnp.log(preds) - (1 - labels) * jnp.log(1 - preds)

    return loss

# %% ../nbs/00_utils.ipynb 48
def sigmoid(x):
    # https://stackoverflow.com/a/68293931
    return 0.5 * (jnp.tanh(x / 2) + 1)

# %% ../nbs/00_utils.ipynb 50
def accuracy(y_true: jnp.ndarray, y_pred: jnp.ndarray) -> jnp.DeviceArray:
    y_true, y_pred = map(jnp.round, (y_true, y_pred))
    return jnp.mean(jnp.equal(y_true, y_pred))


def dist(x: jnp.ndarray, cf: jnp.ndarray, ord: int = 2) -> jnp.DeviceArray:
    dist = jnp.linalg.norm(x - cf, ord=ord, axis=-1, keepdims=True)
    return jnp.mean(vmap(jnp.sum)(dist))


def proximity(x: jnp.ndarray, cf: jnp.ndarray) -> jnp.DeviceArray:
    return dist(x, cf, ord=1)

# %% ../nbs/00_utils.ipynb 53
@dataclass
class Config:
    rng_reserve_size: int
    global_seed: int

    @classmethod
    def default(cls) -> Config:
        return cls(rng_reserve_size=1, global_seed=42)

main_config = Config.default()

# %% ../nbs/00_utils.ipynb 54
def get_config() -> Config: 
    return main_config



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/07_vaecf.ipynb.

# %% ../../nbs/methods/07_vaecf.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from .base import BaseCFModule, BaseParametricCFModule
from ..utils import *
from ..module import MLP, BaseTrainingModule
from ..data import *
from ..trainer import train_model, TrainingConfigs

# %% auto 0
__all__ = ['VAECFConfigs', 'VAECF']

# %% ../../nbs/methods/07_vaecf.ipynb 4
@partial(jax.jit, static_argnums=(3,))
def hindge_embedding_loss(
    inputs: Array, targets: Array, margin: float = 1.0, reduction: str = "mean"
):
    """Hinge embedding loss. (Reduce mean over batch)"""
    assert targets.shape == (1,)
    # assert margin == 1. or margin == -1.
    loss = jnp.where(
        targets == 1,
        inputs,
        jax.nn.relu(margin - inputs)
    )
    if reduction is None:
        return loss
    elif reduction == "mean":
        return jnp.mean(loss)
    elif reduction == "sum":
        return jnp.sum(loss)
    else:
        raise ValueError(f"reduction must be one of [None, 'mean', 'sum'], but got {reduction}")
    # loss = jnp.mean(loss)
    # return loss   


# %% ../../nbs/methods/07_vaecf.ipynb 10
class Encoder(hk.Module):
    def __init__(self, sizes: List[int], dropout: float = 0.1):
        super().__init__()
        self.encoder = MLP(
            sizes[:-1], dropout_rate=dropout, name="encoder_mean")
        self.encoded_size = sizes[-1]
    
    def __call__(self, x: Array, is_training: bool):
        mu = self.encoder(x, is_training)
        mu = hk.Linear(self.encoded_size, name='mu')(mu)
        logvar = self.encoder(x, is_training)
        logvar = hk.Linear(self.encoded_size, name='logvar')(logvar) + 0.5
        logvar = jax.nn.sigmoid(logvar) + 0.5
        return mu, logvar

class Decoder(hk.Module):
    def __init__(
        self, 
        sizes: List[int], 
        input_size: int,
        dropout: float = 0.1
    ):
        super().__init__()
        self.decoder = MLP(
            sizes, dropout_rate=dropout, name="Decoder")
        self.input_size = input_size
    
    def __call__(self, z: Array, is_training: bool):
        mu_dec = self.decoder(z, is_training=is_training)
        mu_dec = hk.Linear(self.input_size, name='mu_x')(mu_dec)
        mu_dec = jax.nn.sigmoid(mu_dec)
        return mu_dec

# %% ../../nbs/methods/07_vaecf.ipynb 11
class VAECFModuleConfigs(BaseParser):
    """Configurator of `VAECFModule`."""
    enc_sizes: List[int] = Field(
        [20, 16, 14, 12, 5],
        description="Sequence of Encoder layer sizes."
    )
    dec_sizes: List[int] = Field(
        [12, 14, 16, 20],
        description="Sequence of Decoder layer sizes."
    )
    dropout_rate: float = Field(
        0.1, description="Dropout rate."
    )
    lr: float = Field(
        1e-3, description="Learning rate."
    )
    mu_samples: int = Field(
        50, description="Number of samples for mu."
    )
    validity_reg: float = Field(
        42.0, description="Regularization for validity."
    )

# %% ../../nbs/methods/07_vaecf.ipynb 12
class VAECFModule(BaseTrainingModule):
    pred_fn: Callable

    def __init__(self, m_configs: Dict = None):
        if m_configs is None: m_configs = {}
        self.save_hyperparameters(m_configs)
        self.m_config = validate_configs(m_configs, VAECFModuleConfigs)
        self.opt = optax.adam(self.m_config.lr)

    def init_net_opt(self, dm, key):
        self._data_module = dm
        X, y = dm.train_dataset[:128]
        Z = jnp.ones((X.shape[0], self.m_config.enc_sizes[-1]))
        inputs = jnp.concatenate([X, y.reshape(-1, 1)], axis=-1)
        latent = jnp.concatenate([Z, y.reshape(-1, 1)], axis=-1)

        self.encoder = make_hk_module(
            Encoder, sizes=self.m_config.enc_sizes, 
            dropout=self.m_config.dropout_rate
        )
        self.decoder = make_hk_module(
            Decoder, sizes=self.m_config.dec_sizes,
            input_size=X.shape[-1], 
            dropout=self.m_config.dropout_rate
        )

        enc_params = self.encoder.init(
            key, inputs, is_training=True)
        dec_params = self.decoder.init(
            key, latent, is_training=True)
        opt_state = self.opt.init((enc_params, dec_params))
        return (enc_params, dec_params), opt_state
    
    @partial(jax.jit, static_argnums=(0, 4))
    def encode(self, enc_params, rng_key, x, is_training=True):
        mu_z, logvar_z = self.encoder.apply(
            enc_params, rng_key, x, is_training=is_training)
        return mu_z, logvar_z
        
    @partial(jax.jit, static_argnums=(0,))
    def sample_latent_code(self, rng_key, mean, logvar):
        eps = jax.random.normal(rng_key, logvar.shape)
        return mean + eps * jnp.sqrt(logvar)
    
    @partial(jax.jit, static_argnums=(0, 4))
    def decode(self, dec_params, rng_key, z, is_training=True):
        mu_x = self.decoder.apply(
            dec_params, rng_key, z, is_training=is_training)
        return mu_x
    
    @partial(jax.jit, static_argnums=(0, 6))
    def sample_step(
        self, rng_key, dec_params, em, ev, c, is_training=True
    ):
        z = self.sample_latent_code(rng_key, em, ev)
        z = jnp.concatenate([z, c.reshape(-1, 1)], axis=-1)
        mu_x = self.decode(dec_params, rng_key, z, is_training=is_training)
        return mu_x
    
    @partial(jax.jit, static_argnums=(0, 4, 5))
    def sample(
        self, params, rng_key, inputs, mc_samples, is_training=True
    ): # Shape: (mc_samples, batch_size, input_size)
        enc_params, dec_params = params
        x, c = inputs[:, :-1], inputs[:, -1]
        em, ev = self.encode(enc_params, rng_key, inputs)
        keys = jax.random.split(rng_key, mc_samples)
        
        partial_sample_step = partial(
            self.sample_step, dec_params=dec_params,
            em=em, ev=ev, c=c
        )
        mu_x = jax.vmap(partial_sample_step)(keys)
        # return SampleOut(em=em, ev=ev, mu_x=mu_x)
        return (em, ev, mu_x)
        
    def compute_loss(self, params, rng_key, inputs, is_training=True):
        def cf_loss(cf: Array, x: Array, y: Array):
            assert cf.shape == x.shape, f"cf.shape ({cf.shape}) != x.shape ({x.shape}))"
            # proximity loss
            recon_err = jnp.sum(jnp.abs(cf - x), axis=1).mean()
            # Sum to 1 over the categorical indexes of a feature
            cat_error = self._data_module.apply_regularization(x, cf)
            # validity loss
            pred_prob = self.pred_fn(cf)
            # This is same as the following:
            # tempt_1, tempt_0 = pred_prob[y == 1], pred_prob[y == 0]
            # validity_loss = hindge_embedding_loss(tempt_1 - (1. - tempt_1), -1, 0.165) + \
            #     hindge_embedding_loss(1. - 2 * tempt_0, -1, 0.165)
            target = jnp.array([-1])
            hindge_loss_1 = hindge_embedding_loss(
                jax.nn.sigmoid(pred_prob) - jax.nn.sigmoid(1. - pred_prob), target, 0.165, reduction=None)
            hindge_loss_0 = hindge_embedding_loss(
                jax.nn.sigmoid(1. - pred_prob) - jax.nn.sigmoid(pred_prob), target, 0.165, reduction=None)
            tempt_1 = jnp.where(y == 1, hindge_loss_1, 0).sum() / y.sum()
            tempt_0 = jnp.where(y == 0, hindge_loss_0, 0).sum() / (y.shape[0] - y.sum())
            validity_loss = tempt_1 + tempt_0

            return recon_err + cat_error, - validity_loss

        em, ev, cfs = self.sample(
            params, rng_key, inputs, self.m_config.mu_samples, 
            is_training=is_training
        )
        X, y = inputs[:, :-1], inputs[:, -1]
        # kl divergence
        kl = 0.5 * jnp.mean(em**2 + ev - jnp.log(ev) - 1, axis=1)
        cf_loss_partial = partial(cf_loss, x=X, y=y)
        recon_err, validity_loss = jax.vmap(cf_loss_partial)(cfs)
        # assert recon_err.shape == (self.m_config.mu_samples,), recon_err.shape
        # assert cat_error.shape == (self.m_config.mu_samples,), cat_error.shape
        # assert validity_loss.shape == (self.m_config.mu_samples,), validity_loss.shape
        recon_err = jnp.mean(recon_err)
        validity_loss = jnp.mean(validity_loss)
        return jnp.mean(kl + recon_err) + validity_loss

    @partial(jax.jit, static_argnums=(0,))
    def _training_step(
        self, 
        params: Tuple[hk.Params, hk.Params],
        opt_state: optax.OptState, 
        rng_key: random.PRNGKey, 
        batch: Tuple[jnp.array, jnp.array]
    ) -> Tuple[hk.Params, optax.OptState]:
        x, _ = batch
        y = self.pred_fn(x).round().reshape(-1, 1)
        loss, grads = jax.value_and_grad(self.compute_loss)(
            params, rng_key, jnp.concatenate([x, y], axis=-1))
        update_params, opt_state = grad_update(
            grads, params, opt_state, self.opt)
        return update_params, opt_state, loss
    
    def training_step(
        self,
        params: Tuple[hk.Params, hk.Params],
        opt_state: optax.OptState,
        rng_key: random.PRNGKey,
        batch: Tuple[jnp.array, jnp.array]
    ) -> Tuple[hk.Params, optax.OptState]:
        params, opt_state, loss = self._training_step(params, opt_state, rng_key, batch)
        self.log_dict({'train/loss': loss.item()})
        return params, opt_state
    
    def validation_step(
        self,
        params: Tuple[hk.Params, hk.Params],
        rng_key: random.PRNGKey,
        batch: Tuple[jnp.array, jnp.array],
    ) -> Tuple[hk.Params, optax.OptState]:
        pass


# %% ../../nbs/methods/07_vaecf.ipynb 13
class VAECFConfigs(VAECFModuleConfigs):
    pass

# %% ../../nbs/methods/07_vaecf.ipynb 14
class VAECF(BaseCFModule, BaseParametricCFModule):
    params: Tuple[hk.Params, hk.Params] = None
    module: VAECFModule
    name: str = 'C-CHVAE'

    def __init__(self, m_config: Dict | VAECFConfigs = None):
        if m_config is None:
            m_config = VAECFConfigs()
        self.m_config = m_config
        self.module = VAECFModule(m_config.dict())

    def _is_module_trained(self) -> bool:
        return not (self.params is None)
    
    def train(
        self, 
        datamodule: TabularDataModule, # data module
        t_configs: TrainingConfigs | dict = None, # training configs
        pred_fn: Callable = None, # prediction function
    ):
        if pred_fn is None:
            raise ValueError('pred_fn must be provided')

        _default_t_configs = dict(
            n_epochs=10, batch_size=128
        )
        if t_configs is None: t_configs = _default_t_configs
        
        setattr(self.module, 'pred_fn', pred_fn)
        params, _ = train_model(self.module, datamodule, t_configs)
        self.params = params
    
    @auto_reshaping('x')
    @partial(jax.jit, static_argnums=[0, 2])
    def generate_cf(
        self, 
        x: Array, 
        pred_fn: Callable = None
    ) -> jnp.ndarray:
        y = pred_fn(x).round().reshape(-1, 1)
        inputs = jnp.concatenate([x, y], axis=-1)
        _, _, cfs = self.module.sample(
            self.params, random.PRNGKey(0), inputs, self.m_config.mu_samples,
            is_training=False
        )
        return self.data_module.apply_constraints(x, cfs[0], hard=True)


    def generate_cfs(self, X: Array, pred_fn: Callable = None) -> jnp.ndarray:
        y = pred_fn(X).round().reshape(-1, 1)
        inputs = jnp.concatenate([X, y], axis=-1)
        _, _, cfs = self.module.sample(
            self.params, random.PRNGKey(0), inputs, self.m_config.mu_samples,
            is_training=False
        )
        return self.data_module.apply_constraints(X, cfs[0], hard=True)
        



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/08_clue.ipynb.

# %% ../../nbs/methods/08_clue.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from .base import BaseCFModule, BaseParametricCFModule
from ..utils import *
from ..module import MLP, BaseTrainingModule
from ..data import *
from ..trainer import train_model, TrainingConfigs
from jax.scipy.stats.norm import logpdf as gaussian_logpdf


# %% auto 0
__all__ = ['Encoder', 'Decoder', 'kl_divergence', 'VAEGaussCatConfigs', 'VAEGaussCat', 'CLUEConfigs', 'CLUE']

# %% ../../nbs/methods/08_clue.ipynb 4
class Encoder(hk.Module):
    def __init__(self, sizes: List[int], dropout: float = 0.1):
        super().__init__()
        assert sizes[-1] % 2 == 0, f"sizes[-1] must be even, but got {sizes[-1]}"
        self.encoder = MLP(
            sizes, dropout_rate=dropout, name="encoder_mean")
    
    def __call__(self, x: Array, is_training: bool):
        params = self.encoder(x, is_training)
        d = params.shape[-1] // 2
        mu, sigma = params[:, :d], params[:, d:]
        sigma = jax.nn.softplus(sigma)
        sigma = jnp.clip(sigma, 1e-3)
        return mu, sigma

class Decoder(hk.Module):
    def __init__(
        self, 
        sizes: List[int], 
        input_size: int,
        dropout: float = 0.1
    ):
        super().__init__()
        self.decoder = MLP(
            sizes, dropout_rate=dropout, name="Decoder")
        self.input_size = input_size
    
    def __call__(self, z: Array, is_training: bool):
        mu_dec = self.decoder(z, is_training=is_training)
        mu_dec = hk.Linear(self.input_size, name='mu_x')(mu_dec)
        mu_dec = jax.nn.sigmoid(mu_dec)
        return mu_dec


# %% ../../nbs/methods/08_clue.ipynb 5
@jit
def kl_divergence(p: Array, q: Array, eps: float = 2 ** -17) -> Array:
    loss_pointwise = p * (jnp.log(p + eps) - jnp.log(q + eps))
    return loss_pointwise

# %% ../../nbs/methods/08_clue.ipynb 6
class VAEGaussCatConfigs(BaseParser):
    lr: float = Field(0.001, description="Learning rate.")
    enc_sizes: List[int] = Field(
        [20, 16, 14, 12],
        description="Sequence of Encoder layer sizes."
    )
    dec_sizes: List[int] = Field(
        [12, 14, 16, 20],
        description="Sequence of Decoder layer sizes."
    )
    dropout_rate: float = Field(
        0.1, description="Dropout rate."
    )



# %% ../../nbs/methods/08_clue.ipynb 7
class VAEGaussCat(BaseTrainingModule):
    def __init__(self, m_configs: Dict = None):
        if m_configs is None: m_configs = {}
        self.save_hyperparameters(m_configs)
        self.m_config = validate_configs(m_configs, VAEGaussCatConfigs)
        self.opt = optax.radam(self.m_config.lr)

    def _update_categorical_info(self):
        cat_arrays = self._data_module._cat_arrays
        self._cat_info = {
            'cat_idx': self._data_module.cat_idx,
            # 'cat_arr': jnp.array([len(cat_arr) for cat_arr in cat_arrays]),
            'cat_arr': [len(cat_arr) for cat_arr in cat_arrays],
        }
    
    def init_net_opt(self, dm, key):
        self._data_module = dm
        self._update_categorical_info()
        keys = jax.random.split(key, 3)
        X, y = dm.train_dataset[:128]
        Z = jnp.ones((X.shape[0], self.m_config.enc_sizes[-1] // 2))

        self.encoder = make_hk_module(
            Encoder, sizes=self.m_config.enc_sizes, 
            dropout=self.m_config.dropout_rate
        )
        self.decoder = make_hk_module(
            Decoder, sizes=self.m_config.dec_sizes,
            input_size=X.shape[-1], 
            dropout=self.m_config.dropout_rate
        )

        enc_params = self.encoder.init(
            keys[0], X, is_training=True)
        dec_params = self.decoder.init(
            keys[1], Z, is_training=True)
        opt_state = self.opt.init((enc_params, dec_params))

        # set prior for training latents
        self.prior = jrand.normal(
            keys[2], (self.m_config.enc_sizes[-1],)
        )
        return (enc_params, dec_params), opt_state
    
    @partial(jax.jit, static_argnums=(0, 4))
    def encode(self, enc_params, rng_key, x, is_training=True):
        mu_z, var_z = self.encoder.apply(
            enc_params, rng_key, x, is_training=is_training)
        return mu_z, var_z
    
    @partial(jax.jit, static_argnums=(0, ))
    def sample_latent(self, rng_key, mean, var):
        key, _ = jax.random.split(rng_key)
        std = jnp.exp(0.5 * var)
        eps = jax.random.normal(key, var.shape)
        return mean + eps * std
    
    @partial(jax.jit, static_argnums=(0, 4))
    def decode(self, dec_params, rng_key, z, is_training=True,):
        reconstruct_x = self.decoder.apply(
            dec_params, rng_key, z, is_training=is_training)
        return reconstruct_x        
    
    @partial(jax.jit, static_argnums=(0, 5))
    def sample_step(
        self, rng_key, dec_params, mean, var, is_training=True
    ):
        z = self.sample_latent(rng_key, mean, var)
        mu_x = self.decode(dec_params, rng_key, z, is_training=is_training)
        return mu_x
    
    @partial(jax.jit, static_argnums=(0, 4, 5))
    def sample(
        self, params, rng_key, x, mc_samples, is_training=True
    ): # Shape: (mc_samples, batch_size, input_size)
        enc_params, dec_params = params
        mean, var = self.encode(enc_params, rng_key, x, is_training=is_training)
        keys = jax.random.split(rng_key, mc_samples)
        
        partial_sample_step = partial(
            self.sample_step, dec_params=dec_params,
            mean=mean, var=var, is_training=is_training
        )
        reconstruct_x = jax.vmap(partial_sample_step)(keys)
        return (mean, var, reconstruct_x)
    
    @partial(jax.jit, static_argnums=(0, ))
    def sample_prior(self, rng_key):
        rng_key, key = jax.random.split(rng_key)
        prior = jrand.normal(key, (self.m_config.enc_sizes[-1],))
        return prior
    
    def compute_loss(self, params, rng_key, x, is_training=True):
        # @partial(jax.jit, static_argnums=(2, 3))
        def reconstruct_loss(x: Array, cf: Array, cat_idx: int, cat_arr: List[int]):
            cont_loss = optax.l2_loss(x[:, :cat_idx], cf[:, :cat_idx])
            cat_loss = []

            def _cat_loss_f(start_end_idx):
                start_idx, end_idx = start_end_idx
                return optax.softmax_cross_entropy(
                    cf[:, start_idx: end_idx], x[:, start_idx: end_idx]
                ).reshape(-1, 1)
            
            # for start_end_idx in start_end_indices:
            for i, cat in enumerate(cat_arr):
                start_end_idx = (cat_idx + i * cat, cat_idx + (i + 1) * cat)
                cat_loss.append(_cat_loss_f(start_end_idx))
            cat_loss = jnp.concatenate(cat_loss, axis=-1)
            
            # cat_loss = jax.vmap(jit(_cat_loss_f))(start_indices, end_indices)
            # cat_loss = jax.lax.scan(_cat_loss_f, 0., start_end_indices, len(start_end_indices))[1]
            return jnp.concatenate([cont_loss, cat_loss], axis=-1).sum(-1)
        
        keys = jax.random.split(rng_key, 2)
        mu_z, logvar_z, reconstruct_x = self.sample(
            params, keys[0], x, mc_samples=1, is_training=is_training
        )
        kl_loss = -0.5 * (1 + logvar_z - jnp.power(mu_z, 2) - jnp.exp(logvar_z)).sum(-1)
        
        rec = reconstruct_loss(
            x, reconstruct_x.reshape(x.shape), 
            cat_idx=self._cat_info['cat_idx'],
            cat_arr=self._cat_info['cat_arr']
        ).sum(-1)
        batchwise_loss = (rec + kl_loss) / x.shape[0]
        return batchwise_loss.mean()

    @partial(jax.jit, static_argnums=(0,))
    def _training_step(
        self, 
        params: Tuple[hk.Params, hk.Params],
        opt_state: optax.OptState, 
        rng_key: random.PRNGKey, 
        batch: Tuple[Array, Array]
    ) -> Tuple[hk.Params, optax.OptState]:
        x, _ = batch
        loss, grads = jax.value_and_grad(self.compute_loss)(
            params, rng_key, x)
        update_params, opt_state = grad_update(
            grads, params, opt_state, self.opt)
        return update_params, opt_state, loss

    def training_step(
        self,
        params: Tuple[hk.Params, hk.Params],
        opt_state: optax.OptState,
        rng_key: random.PRNGKey,
        batch: Tuple[jnp.array, jnp.array]
    ) -> Tuple[hk.Params, optax.OptState]:
        params, opt_state, loss = self._training_step(params, opt_state, rng_key, batch)
        self.log_dict({'train/loss': loss.item()})
        return params, opt_state
    
    @partial(jax.jit, static_argnums=(0,))
    def validation_step(
        self,
        params: Tuple[hk.Params, hk.Params],
        rng_key: random.PRNGKey,
        batch: Tuple[jnp.array, jnp.array],
    ) -> Tuple[hk.Params, optax.OptState]:
        pass


# %% ../../nbs/methods/08_clue.ipynb 8
@auto_reshaping('x')
def _clue_generate(
    x: Array,
    rng_key: jrand.PRNGKey,
    pred_fn: Callable,
    max_steps: int,
    step_size: float,
    vae_module: VAEGaussCat,
    vae_params: Tuple[hk.Params, hk.Params],
    uncertainty_weight: float,
    aleatoric_weight: float,
    prior_weight: float,
    distance_weight: float,
    validity_weight: float,
    apply_fn: Callable
) -> Array:
    
    @jit
    def sample_latent_from_x(
        x: Array, enc_params: hk.Params, rng_key: jrand.PRNGKey
    ):
        key_1, key_2 = jrand.split(rng_key)
        mean, var = vae_module.encode(enc_params, key_1, x, is_training=False)
        z = vae_module.sample_latent(key_2, mean, var)
        return z
    
    @partial(jit, static_argnums=(2,))
    def generate_from_z(
        z: Array, 
        dec_params: hk.Params,
        hard: bool = False
    ):
        cf = vae_module.decode(
            dec_params, rng_key, z, is_training=False)
        cf = apply_fn(x, cf, hard=hard)
        return cf

    @jit
    def uncertainty_from_z(z: Array, dec_params: hk.Params):
        cfs = generate_from_z(z, dec_params, hard=False)
        prob = pred_fn(cfs)
        total_uncertainty = -(prob * jnp.log(prob + 1e-10)).sum(-1)
        return total_uncertainty, cfs, prob
    
    @jit
    def compute_loss(z: Array, dec_params: hk.Params):
        uncertainty, cfs, prob = uncertainty_from_z(z, dec_params)
        loglik = gaussian_logpdf(z).sum(-1)
        dist = jnp.abs(cfs - x).mean()
        validity = binary_cross_entropy(preds=prob, labels=y_targets).mean()
        loss = (
            (uncertainty_weight + aleatoric_weight) * uncertainty 
            + prior_weight * loglik
            + distance_weight * dist
            + validity_weight * validity
        )
        return loss.mean()
    
    def step(i, z_opt_state):
        z, opt_state = z_opt_state
        z_grad = jax.grad(compute_loss)(z, dec_params)
        z, opt_state = grad_update(z_grad, z, opt_state, opt)
        return z, opt_state
    
    enc_params, dec_params = vae_params
    key_1, _ = jax.random.split(rng_key)
    z = sample_latent_from_x(x, enc_params, key_1)
    opt = optax.adam(step_size)
    opt_state = opt.init(z)
    y_targets = 1 - pred_fn(x)

    # Write a loop to optimize z using lax.fori_loop
    z, opt_state = lax.fori_loop(0, max_steps, step, (z, opt_state))
    cf = generate_from_z(z, dec_params, hard=True)
    return cf


# %% ../../nbs/methods/08_clue.ipynb 9
class CLUEConfigs(BaseParser):
    enc_sizes: List[int] = Field(
        [20, 16, 14, 12], description="Sequence of Encoder layer sizes."
    )
    dec_sizes: List[int] = Field(
        [12, 14, 16, 20], description="Sequence of Decoder layer sizes."
    )
    encoded_size: int = Field(5, description="Encoded size")
    lr: float = Field(0.001, description="Learning rate")
    max_steps: int = Field(500, description="Max steps")
    step_size: float = Field(0.01, description="Step size")
    vae_n_epochs: int = Field(10, description="Number of epochs for VAE")
    vae_batch_size: int = Field(128, description="Batch size for VAE")
    seed: int = Field(0, description="Seed for random number generator")

# %% ../../nbs/methods/08_clue.ipynb 10
class CLUE(BaseCFModule, BaseParametricCFModule):
    params: Tuple[hk.Params, hk.Params] = None
    module: VAEGaussCat
    name: str = 'CLUE'

    def __init__(self, m_config: Dict | CLUEConfigs = None):
        if m_config is None: m_config = CLUEConfigs()
        self.m_config = m_config
        self.module = VAEGaussCat(m_config.dict())
        self.rng_key = random.PRNGKey(self.m_config.seed)

    def _is_module_trained(self) -> bool:
        return not (self.params is None)
    
    def train(
        self, 
        datamodule: TabularDataModule, # data module
        t_configs: TrainingConfigs | dict = None, # training configs
        *args, **kwargs
    ):
        _default_t_configs = dict(
            n_epochs=10, batch_size=128
        )
        if t_configs is None: t_configs = _default_t_configs
        params, _ = train_model(self.module, datamodule, t_configs)
        self.params = params

    def generate_cf(self, x, pred_fn: Callable = None) -> Array:
        return _clue_generate(
            x, 
            rng_key=self.rng_key, 
            pred_fn=pred_fn,
            max_steps=self.m_config.max_steps,
            step_size=self.m_config.step_size,
            vae_module=self.module,
            vae_params=self.params,
            uncertainty_weight=.0,
            aleatoric_weight=0.0,
            prior_weight=0.0,
            distance_weight=.1,
            validity_weight=1.0,
            apply_fn=self.data_module.apply_constraints,
        )
    
    def generate_cfs(self, X: Array, pred_fn: Callable = None) -> jnp.ndarray:
        generate_cf_partial = partial(
            self.generate_cf, pred_fn=pred_fn
        )
        rngs = lax.broadcast(random.PRNGKey(0), (X.shape[0], ))
        return jax.vmap(generate_cf_partial)(X, rngs)
